# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import typing

from opteryx import config
from opteryx.connectors import BaseDocumentStorageAdapter
from opteryx.connectors.capabilities import PredicatePushable
from opteryx.exceptions import MissingDependencyError
from opteryx.exceptions import UnmetRequirementError

GCP_PROJECT_ID = config.GCP_PROJECT_ID


def _get_project_id():  # pragma: no cover
    """Fetch the ID from GCP"""
    try:
        import requests
    except ImportError as exception:  # pragma: no cover
        raise UnmetRequirementError(
            "Firestore requires 'GCP_PROJECT_ID` to be set in config, or "
            "`requests` to be installed."
        ) from exception

    # if it's set in the config/environ, use that
    if GCP_PROJECT_ID:
        return GCP_PROJECT_ID

    # otherwise try to get it from GCP
    response = requests.get(
        "http://metadata.google.internal/computeMetadata/v1/project/project-id",
        headers={"Metadata-Flavor": "Google"},
        timeout=10,
    )
    response.raise_for_status()
    return response.text


def _initialize():  # pragma: no cover
    """Create the connection to Firebase"""
    try:
        import firebase_admin
        from firebase_admin import credentials
    except ImportError as err:  # pragma: no cover
        raise MissingDependencyError(err.name) from err
    if not firebase_admin._apps:
        # if we've not been given the ID, fetch it
        project_id = GCP_PROJECT_ID
        if project_id is None:
            project_id = _get_project_id()
        creds = credentials.ApplicationDefault()
        firebase_admin.initialize_app(creds, {"projectId": project_id})


class GcpFireStoreConnector(BaseDocumentStorageAdapter, PredicatePushable):
    def __init__(self, *args, prefix: str = "", remove_prefix: bool = False, **kwargs):
        BaseDocumentStorageAdapter.__init__(
            self, *args, prefix=prefix, remove_prefix=remove_prefix, **kwargs  # type: ignore
        )
        PredicatePushable.__init__(
            self, *args, prefix=prefix, remove_prefix=remove_prefix, **kwargs
        )
        self.supported_ops = ["=="]

        self._remove_prefix = remove_prefix
        self._prefix = prefix

    def get_document_count(self, collection) -> int:  # pragma: no cover
        """
        Return an interable of blobs/files

        FireStore currently doesn't support this
        """
        return -1

    def read_documents(self, collection, morsel_size: typing.Union[int, None] = None):
        """
        Return a morsel of documents
        """
        from firebase_admin import firestore

        if morsel_size is None:
            morsel_size = config.MORSEL_SIZE

        queried_collection = collection
        if self._remove_prefix:
            if collection.startswith(f"{self._prefix}."):
                queried_collection = collection[len(self._prefix) + 1 :]

        _initialize()
        database = firestore.client()
        documents = database.collection(queried_collection)

        for predicate in self._predicates:
            documents = documents.where(*predicate)

        documents = documents.stream()

        for morsel in self.chunk_dictset(
            ({**doc.to_dict(), "_id": doc.id} for doc in documents), morsel_size
        ):
            yield morsel



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from opteryx import config
from opteryx.managers.kvstores import HadroDB
from opteryx.managers.kvstores import KV_store_factory


def metadata_factory():
    if config.METADATA_SERVER is None or config.METADATA_SERVER.upper() == "LOCAL":
        # HadroDB is our default KV store
        return HadroDB
    else:
        # use the configured one
        return KV_store_factory(config.METADATA_SERVER)


def put():
    pass


def get():
    pass



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Rewrite Rule - Literal Join Filters

Goal: Move filters JOIN which reference literals
"""
from opteryx import operators
from opteryx.managers.expression import LITERAL_TYPE
from opteryx.managers.expression import NodeType
from opteryx.utils import random_string


def move_literal_join_filters(plan, properties):
    """
    This action aims to move any filters in JOIN ON conditions which are referencing
    literals and therefore aren't actually a join condition.

    If the condition doesn't have a literal, then we assume it's a join condition.

    When we remove the filters, we put them after the join - this is always correct
    but non-optimal, that's fine we should let the optimizer work out where to filter.
    """

    def _has_literal(node):
        return (node.left and node.left.token_type & LITERAL_TYPE == LITERAL_TYPE) or (
            node.right and node.right.token_type & LITERAL_TYPE == LITERAL_TYPE
        )

    def _move_literal_filters(plan, nid, node):
        # we need to be at an AND
        if node.token_type != NodeType.AND:
            return node, plan

        uid = random_string()  # avoid collisions
        if _has_literal(node.left):
            # we create a new selection node with the literal filter and add it to the plan
            new_node = operators.SelectionNode(filter=node.left, properties=properties)
            plan.insert_node_after(f"select-{nid}-{uid}-left", new_node, nid)
            # we remove the filter from this node
            node = node.right
        elif _has_literal(node.right):
            new_node = operators.SelectionNode(filter=node.right, properties=properties)
            plan.insert_node_after(f"select-{nid}-{uid}-right", new_node, nid)
            # we remove the filter from this node
            node = node.left

        return node, plan

    # find the in-scope nodes (INNER and OUTER joins)
    join_nodes = plan.get_nodes_of_type([operators.InnerJoinNode, operators.OuterJoinNode])

    # HAVING and WHERE are selection nodes
    for nid in join_nodes:
        # get the node from the node_id
        operator = plan[nid]
        if operator._on is not None:
            operator._on, plan = _move_literal_filters(plan, nid, operator._on)

    return plan



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module provides a PEP-249 familiar interface for interacting with mabel data
stores, it is not compliant with the standard:
https://www.python.org/dev/peps/pep-0249/
"""
import logging
import time
import typing
from uuid import uuid4

import pyarrow
from orso import DataFrame
from orso import converters

from opteryx import utils
from opteryx.exceptions import CursorInvalidStateError
from opteryx.exceptions import MissingSqlStatement
from opteryx.exceptions import PermissionsError
from opteryx.exceptions import ProgrammingError
from opteryx.managers.kvstores import BaseKeyValueStore
from opteryx.shared import QueryStatistics

CURSOR_NOT_RUN: str = "Cursor must be in an executed state"


class Connection:
    """
    A connection
    """

    def __init__(
        self,
        *,
        cache: typing.Union[BaseKeyValueStore, None] = None,
        permissions: typing.Union[typing.Iterable, None] = None,
        **kwargs,
    ):
        """
        A virtual connection to the Opteryx query engine.
        """
        self._results = None
        self.cache = cache
        self._kwargs = kwargs

        self._default_database = None
        self._connection_id = utils.random_int()

        # check the permissions we've been given are valid permissions
        from opteryx.constants.permissions import PERMISSIONS

        if permissions is None:
            permissions = PERMISSIONS
        permissions = set(permissions)
        if permissions.intersection(PERMISSIONS) == set():
            raise ProgrammingError("No valid permissions presented.")
        if not permissions.issubset(PERMISSIONS):
            raise ProgrammingError(
                f"Invalid permissions presented - {PERMISSIONS.difference(permissions)}"
            )
        self.permissions = permissions

    def cursor(self):
        """return a cursor object"""
        return Cursor(self)

    def close(self):
        """exists for interface compatibility only"""
        pass

    def commit(self):
        """exists for interface compatibility only"""
        pass

    def rollback(self):
        """exists for interface compatibility only"""
        # return AttributeError as per https://peps.python.org/pep-0249/#id48
        raise AttributeError("Opteryx does not support transactions.")


class Cursor(DataFrame):
    def __init__(self, connection):
        self.arraysize = 1
        self._connection = connection
        self._query = None
        self._query_planner = None
        self._collected_stats = None
        self._plan = None
        self._qid = str(uuid4())
        self._statistics = QueryStatistics(self._qid)
        DataFrame.__init__(self, rows=[], schema=[])

    @property
    def query(self):
        return self._query

    @property
    def id(self):
        """The unique internal reference for this query"""
        return self._qid

    def _inner_execute(self, operation, params=None):
        if not operation:
            raise MissingSqlStatement("SQL statement not found")

        if self._query is not None:
            raise CursorInvalidStateError("Cursor can only be executed once")
        self._query = operation

        from opteryx.components.query_planner import QueryPlanner

        self._query_planner = QueryPlanner(
            statement=operation, cache=self._connection.cache, qid=self._qid
        )
        self._statistics.start_time = time.time_ns()
        asts = list(self._query_planner.parse_and_lex())

        # test permissions
        for ast in asts:
            statement_type = next(iter(ast))
            if statement_type not in self._connection.permissions:
                raise PermissionsError(f"Required permission {statement_type} not provided.")

        results = None
        if params is None:
            params = []

        self._query_planner.test_paramcount(asts, params)

        for ast in asts:
            ast = self._query_planner.bind_ast(ast, parameters=params)
            plan = self._query_planner.create_logical_plan(ast)

            self._plan = self._query_planner.optimize_plan(plan)
            results = self._query_planner.execute(self._plan)

        if results is not None:
            return utils.arrow.rename_columns(results)

    def execute(self, operation, params=None):
        results = self._inner_execute(operation, params)
        if results is not None:
            self._rows, self._schema = converters.from_arrow(results)
            self._cursor = iter(self._rows)

    def execute_to_arrow(self, operation, params=None, limit=None):
        results = self._inner_execute(operation, params)
        if results is not None:
            if limit is not None:
                return utils.arrow.limit_records(results, limit)
        return pyarrow.concat_tables(results, promote=True)

    @property
    def stats(self):
        """execution statistics"""
        if self._statistics.end_time == 0:  # pragma: no cover
            self._statistics.end_time = time.time_ns()
        return self._statistics.as_dict()

    @property
    def messages(self) -> list:
        """list of run-time warnings"""
        return self._statistics.messages

    def to_df(self, size: int = None):  # pragma: no cover
        logging.warning("connection.to_df will be removed, replace with connection.pandas")
        return self.pandas(size)

    def close(self):
        """close the connection"""
        self._connection.close()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from hadrodb import HadroDB

from .base_kv_store import BaseKeyValueStore
from .kv_firestore import FireStoreKVStore


def KV_store_factory(store):  # pragma: no-cover
    """
    A factory method for getting KV Store instances
    """
    stores = {
        "hadro": HadroDB,
        "firestore": FireStoreKVStore,
    }

    return stores.get(store.lower(), HadroDB)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# fmt: off
from .rule_apply_demorgans_law import apply_demorgans_law
from .rule_constant_evaluations import eliminate_constant_evaluations
from .rule_defragment_morsels import defragment_morsels
from .rule_eliminate_negations import eliminate_negations
from .rule_function_evaluations import eliminate_fixed_function_evaluations
from .rule_move_literal_join_filters import move_literal_join_filters
from .rule_predicate_pushdown import predicate_pushdown
from .rule_split_conjunctive_predicates import split_conjunctive_predicates
from .rule_use_heap_sort import use_heap_sort

# fmt:on



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Base Inner Reader for nosql document stores
"""
import abc
from typing import Iterable

import pyarrow


class BaseDocumentStorageAdapter(abc.ABC):
    __mode__ = "Collection"

    def __init__(self, prefix: str, remove_prefix: bool = False) -> None:
        self._prefix = prefix
        self._remove_prefix = remove_prefix
        self.chunk_size = 500

    def chunk_dictset(self, dictset: Iterable[dict], morsel_size: int):
        """
        Enables paging through a dictset by returning a chunk of records at a time.
        Parameters:
            dictset: iterable of dictionaries:
                The dictset to process
            chunk_size: integer:
                The number of bytes per chunk
        """
        chunk_size = self.chunk_size
        index = -1
        chunk: list = [{}] * chunk_size
        for index, record in enumerate(dictset):
            _id = record.pop("_id", None)
            record["id"] = None if _id is None else str(_id)
            if index > 0 and index % chunk_size == 0:
                morsel = pyarrow.Table.from_pylist(chunk)
                # from 500 records, estimate the number of records to fill the morsel size
                # in the unlikely event that the right size for chunks is 500, we calculate
                # this every cycle
                if chunk_size == 500 and morsel.nbytes > 0:
                    chunk_size = int(morsel_size // (morsel.nbytes / 500))
                    self.chunk_size = chunk_size
                # yield after the calculation, as you probably expect this to be set after
                # the first read, not the second read
                yield morsel
                chunk = [{}] * chunk_size
                chunk[0] = record
            else:
                chunk[index % chunk_size] = record
        morsel = pyarrow.Table.from_pylist(chunk[: (index + 1) % chunk_size])
        yield morsel

    def get_document_count(self, collection) -> int:  # pragma: no cover
        """
        Return the count, or an estimate of, the number of documents
        """
        raise NotImplementedError("get_document_list not implemented")

    def read_documents(self, collection, morsel_size: int = 500):  # pragma: no cover
        """
        Return a morsel of documents
        """
        raise NotImplementedError("read_document not implemented")

    @property
    def can_push_selection(self):
        return False



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Left Join Node

This is a SQL Query Execution Plan Node.

This performs a LEFT (OUTER) JOIN
"""
from typing import Iterable

import numpy
import pyarrow

from opteryx.exceptions import ColumnNotFoundError
from opteryx.exceptions import SqlError
from opteryx.models import Columns
from opteryx.models import QueryProperties
from opteryx.operators import BasePlanNode
from opteryx.third_party import pyarrow_ops
from opteryx.utils import arrow

OUTER_JOINS = {
    "FullOuter": "Full Outer",
    "LeftOuter": "Left Outer",
    "RightOuter": "Right Outer",
}

INTERNAL_BATCH_SIZE = 500  # config


def calculate_batch_size(cardinality):
    """dynamically work out the processing batch size for the USING JOIN"""
    # - HIGH_CARDINALITY_BATCH_SIZE (over 90% unique) = INTERNAL_BATCH_SIZE
    # - MEDIUM_CARDINALITY_BATCH_SIZE (5% > n < 90%) = INTERNAL_BATCH_SIZE * n
    # - LOW_CARDINALITY_BATCH_SIZE (less than 5% unique) = 5
    # These numbers have had very little science put into them, they are unlikely
    # to be optimal
    if cardinality < 0.05:
        return 5
    if cardinality > 0.9:
        return INTERNAL_BATCH_SIZE
    return INTERNAL_BATCH_SIZE * cardinality


class OuterJoinNode(BasePlanNode):
    def __init__(self, properties: QueryProperties, **config):
        super().__init__(properties=properties)
        self._join_type = OUTER_JOINS[config.get("join_type")]
        self._on = config.get("join_on")
        self._using = config.get("join_using")

        if self._on is None and self._using is None:
            raise SqlError("Missing JOIN 'ON' condition.")

        if self._using is not None and self._join_type.lower() != "left outer":
            raise SqlError("JOIN `USING` only valid for `INNER` and `LEFT` joins.")

    @property
    def name(self):  # pragma: no cover
        return f"{self._join_type} Join"

    @property
    def config(self):  # pragma: no cover
        return ""

    def execute(self) -> Iterable:
        if len(self._producers) != 2:  # pragma: no cover
            raise SqlError(f"{self.name} expects two producers")

        left_node = self._producers[0]  # type:ignore
        right_node = self._producers[1]  # type:ignore

        right_table = pyarrow.concat_tables(right_node.execute(), promote=True)

        right_columns = Columns(right_table)
        left_columns = None

        for morsel in left_node.execute():
            if self._using:
                right_join_columns = [
                    right_columns.get_column_from_alias(col, only_one=True) for col in self._using
                ]

                for morsel in left_node.execute():
                    if left_columns is None:
                        left_columns = Columns(morsel)
                        left_join_columns = [
                            left_columns.get_column_from_alias(col, only_one=True)
                            for col in self._using
                        ]
                        new_metadata = left_columns + right_columns

                        # we estimate the cardinality of the left table to inform the
                        # batch size for the joins. Cardinality here is the ratio of
                        # unique values in the set. Although we're working it out, we'll
                        # refer to this as an estimate because it may be different per
                        # chunk of data - we're assuming it's not very different.
                        cols = pyarrow_ops.columns_to_array_denulled(morsel, left_join_columns)
                        if morsel.num_rows > 0:
                            card = len(numpy.unique(cols)) / morsel.num_rows
                        else:
                            card = 0
                        batch_size = calculate_batch_size(card)

                    # we break this into small chunks otherwise we very quickly run into memory issues
                    for batch in morsel.to_batches(max_chunksize=batch_size):
                        # blocks don't have column_names, so we need to wrap in a table
                        batch = pyarrow.Table.from_batches([batch], schema=morsel.schema)

                        new_morsel = pyarrow_ops.left_join(
                            right_table, batch, right_join_columns, left_join_columns
                        )
                        new_morsel = new_metadata.apply(new_morsel)
                        yield new_morsel

            elif self._on:
                if left_columns is None:
                    left_columns = Columns(morsel)
                    try:
                        right_join_column = right_columns.get_column_from_alias(
                            self._on.right.value, only_one=True
                        )
                        left_join_column = left_columns.get_column_from_alias(
                            self._on.left.value, only_one=True
                        )
                    except ColumnNotFoundError:
                        # the ON condition may not always be in the order of the tables
                        # we purposefully reference the columns incorrectly
                        right_join_column = right_columns.get_column_from_alias(
                            self._on.left.value, only_one=True
                        )
                        left_join_column = left_columns.get_column_from_alias(
                            self._on.right.value, only_one=True
                        )

                    # ensure the types are compatible for joining by coercing numerics
                    right_table = arrow.coerce_columns(right_table, right_join_column)

                new_metadata = right_columns + left_columns

                morsel = arrow.coerce_columns(morsel, left_join_column)

                # do the join
                new_morsel = morsel.join(
                    right_table,
                    keys=[left_join_column],
                    right_keys=[right_join_column],
                    join_type=self._join_type.lower(),
                    coalesce_keys=False,
                )
                # update the metadata
                new_morsel = new_metadata.apply(new_morsel)
                yield new_morsel



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Builders which require special handling
"""
import datetime
from dataclasses import dataclass
from dataclasses import field
from typing import Any

from opteryx.managers.expression import ExpressionTreeNode
from opteryx.managers.expression import NodeType
from opteryx.shared import QueryStatistics
from opteryx.utils import fuzzy_search
from opteryx.utils import paths

WELL_KNOWN_HINTS = {
    "NO_CACHE",
    "NO_PARTITION",
    "NO_PUSH_PROJECTION",
    "NO_PUSH_SELECTION",
}


@dataclass
class RelationDescription:
    dataset: str = None
    alias: str = None
    kind: str = None
    hints: list = field(default_factory=list)
    start_date: datetime.date = None
    end_date: datetime.date = None
    cache: Any = None


def extract_show_filter(ast, name_column="name"):
    """filters are used in SHOW queries"""
    filters = ast["filter"]
    if filters is None:
        return None
    if "Like" in filters:
        left = ExpressionTreeNode(NodeType.IDENTIFIER, value=name_column)
        right = ExpressionTreeNode(NodeType.LITERAL_VARCHAR, value=filters["Like"])
        root = ExpressionTreeNode(
            NodeType.COMPARISON_OPERATOR,
            value="ILike",  # we're case insensitive for SHOW filters
            left=left,
            right=right,
        )
        return root
    if "Where" in filters:
        from opteryx.components.logical_planner import builders

        root = builders.build(filters["Where"])
        return root


def extract_distinct(ast):
    return ast["Query"]["body"]["Select"]["distinct"]


def extract_limit(ast):
    limit = ast["Query"].get("limit")
    if limit is not None:
        return int(limit["Value"]["Number"][0])
    return None


def extract_offset(ast):
    offset = ast["Query"].get("offset")
    if offset is not None:
        return int(offset["value"]["Value"]["Number"][0])
    return None


def extract_order(ast):
    from opteryx.components.logical_planner import builders

    order = ast["Query"].get("order_by")
    if order is not None:
        orders = []
        for col in order:
            column = builders.build([col["expr"]])
            orders.append(
                (
                    column,
                    "descending" if str(col["asc"]) == "False" else "ascending",
                ),
            )
        return orders


def extract_identifiers(ast):
    identifiers = []
    if ast == {
        "name": [{"value": "COUNT", "quote_style": None}],
        "args": [{"Unnamed": "Wildcard"}],
        "over": None,
        "distinct": False,
        "special": False,
    } or ast == {
        "name": [{"value": "count", "quote_style": None}],
        "args": [{"Unnamed": "Wildcard"}],
        "over": None,
        "distinct": False,
        "special": False,
    }:
        identifiers.append("count_*")
    elif isinstance(ast, dict):
        for key, value in ast.items():
            if key in ("Identifier",):
                identifiers.append(value["value"])
            if key in ("Using",):
                for item in ast["Using"]:
                    identifiers.append(item["value"])
            if key in ("QualifiedWildcard",):
                identifiers.append("*")
            identifiers.extend(extract_identifiers(value))
    elif isinstance(ast, list):
        for item in ast:
            if isinstance(item, dict):
                _key = next(iter(item))
                if _key in ("Wildcard",):
                    identifiers.append("*")
            elif item == "Wildcard":
                identifiers.append("*")
            identifiers.extend(extract_identifiers(item))

    return identifiers


def extract_joins(ast, qid):
    from opteryx.components.logical_planner import builders

    try:
        joins = ast["Query"]["body"]["Select"]["from"][0]["joins"]
    except IndexError:
        return None

    for join in joins:
        join_using = None
        join_on = None
        join_mode = join["join_operator"]
        if isinstance(join_mode, dict):
            join_mode = list(join["join_operator"].keys())[0]
            if "Using" in join["join_operator"][join_mode]:
                join_using = [v["value"] for v in join["join_operator"][join_mode].get("Using", [])]
            if "On" in join["join_operator"][join_mode]:
                join_on = builders.build(join["join_operator"][join_mode]["On"])

        right = next(extract_relations([join], qid))
        yield (join_mode, right, join_on, join_using)


def extract_relations(branch, qid):
    """ """
    from opteryx.components.logical_planner import builders
    from opteryx.components.query_planner import QueryPlanner

    def _check_hints(hints):
        for hint in hints:
            if hint not in WELL_KNOWN_HINTS:
                best_match_hint = fuzzy_search(hint, WELL_KNOWN_HINTS)
                statistics = QueryStatistics(qid)
                if best_match_hint:
                    statistics.add_message(
                        f"Hint `{hint}` is not recognized, did you mean `{best_match_hint}`?"
                    )
                else:
                    statistics.add_message(f"Hint `{hint}` is not recognized.")

    for relation in branch:
        relation_desc = RelationDescription()
        if "Table" in relation["relation"]:
            # is the relation a builder function
            if relation["relation"]["Table"]["args"]:
                function = relation["relation"]["Table"]["name"][0]["value"].lower()
                relation_desc.alias = function
                if relation["relation"]["Table"]["alias"] is not None:
                    relation_desc.alias = relation["relation"]["Table"]["alias"]["name"]["value"]
                args = [builders.build(a["Unnamed"]) for a in relation["relation"]["Table"]["args"]]
                relation_desc.kind = "Function"
                relation_desc.dataset = {
                    "function": function,
                    "args": args,
                }  # type:ignore
                yield relation_desc
            else:
                if relation["relation"]["Table"]["alias"] is not None:
                    relation_desc.alias = relation["relation"]["Table"]["alias"]["name"]["value"]
                if relation["relation"]["Table"]["with_hints"] is not None:
                    relation_desc.hints = [
                        hint["Identifier"]["value"]
                        for hint in relation["relation"]["Table"]["with_hints"]
                    ]
                    # hint checks
                    _check_hints(relation_desc.hints)
                relation_desc.dataset = ".".join(
                    [part["value"] for part in relation["relation"]["Table"]["name"]]
                )

                relation_desc.start_date = relation["relation"]["Table"]["start_date"]
                relation_desc.end_date = relation["relation"]["Table"]["end_date"]
                relation_desc.cache = relation["relation"]["Table"]["cache"]
                if relation_desc.dataset[0:1] == "$":
                    relation_desc.kind = "Internal"
                elif relation_desc.dataset.startswith("information_schema"):
                    relation_desc.kind = "InformationSchema"
                elif paths.is_file(relation_desc.dataset):
                    relation_desc.kind = "File"
                else:
                    relation_desc.kind = "External"
                yield relation_desc

        if "Derived" in relation["relation"]:
            subquery = relation["relation"]["Derived"]["subquery"]["body"]
            try:
                relation_desc.alias = relation["relation"]["Derived"]["alias"]["name"]["value"]
            except (KeyError, TypeError):
                pass
            if "Select" in subquery:
                ast = {}
                ast["Query"] = relation["relation"]["Derived"]["subquery"]

                subquery_planner = QueryPlanner(qid=qid)
                plan = subquery_planner.create_logical_plan(ast)
                plan = subquery_planner.optimize_plan(plan)

                relation_desc.dataset = plan
                relation_desc.kind = "SubQuery"
                yield relation_desc
            if "Values" in subquery:
                body = []
                headers = [h["value"] for h in relation["relation"]["Derived"]["alias"]["columns"]]
                for value_set in subquery["Values"]["rows"]:
                    values = [builders.build(v["Value"]).value for v in value_set]
                    body.append(dict(zip(headers, values)))
                relation_desc.dataset = {
                    "function": "values",
                    "args": body,
                }  # type:ignore
                relation_desc.kind = "Function"
                yield relation_desc


def extract_into(branch):
    return None


def extract_ctes(branch, qid):
    from opteryx.components.query_planner import QueryPlanner

    ctes = {}

    if branch["with"]:
        for ast in branch["with"]["cte_tables"]:
            subquery_planner = QueryPlanner(qid=qid)
            alias = ast.pop("alias")["name"]["value"]
            plan = subquery_planner.create_logical_plan({"Query": ast["query"]})
            ctes[alias] = plan

    return ctes



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Optimization Rule - Eliminate Constant Evaluations

Type: Heuristic
Goal: Reduce complexity
"""
import datetime
import decimal

import numpy

from opteryx import operators
from opteryx.functions.binary_operators import binary_operations
from opteryx.managers.expression import LITERAL_TYPE
from opteryx.managers.expression import ExpressionTreeNode
from opteryx.managers.expression import NodeType
from opteryx.third_party.pyarrow_ops.ops import filter_operations


def eliminate_constant_evaluations(plan, properties):
    """
    This action aims to remove repeat evaluations of constant expressions.
    """

    def build_literal_node(value):
        # fmt:off
        if hasattr(value, "as_py"):
            value = value.as_py()
        if value is None:
            return ExpressionTreeNode(NodeType.LITERAL_NONE, alias=[])
        if isinstance(value, (bool, numpy.bool_)):
            # boolean must be before numeric
            return ExpressionTreeNode(NodeType.LITERAL_BOOLEAN, value=value, alias=[])
        if isinstance(value, (str)):
            return ExpressionTreeNode(NodeType.LITERAL_VARCHAR, value=value, alias=[])
        if isinstance(value, (int, float, decimal.Decimal)):
            return ExpressionTreeNode(NodeType.LITERAL_NUMERIC, value=value, alias=[])
        if isinstance(value, (numpy.datetime64)):
            return ExpressionTreeNode(NodeType.LITERAL_TIMESTAMP, value=value, alias=[])
        if isinstance(value, (datetime.date, datetime.datetime)):
            return ExpressionTreeNode(NodeType.LITERAL_TIMESTAMP, value=numpy.datetime64(value), alias=[])
        # fmt:on

    def update_expression_tree(node):
        """
        Walk a expression tree collecting all the nodes of a specified type.
        """
        # walk the tree first so we can bubble constants as far up the tree as possible
        node.left = None if node.left is None else update_expression_tree(node.left)
        node.centre = None if node.centre is None else update_expression_tree(node.centre)
        node.right = None if node.right is None else update_expression_tree(node.right)

        if node.parameters:
            node.parameters = [
                parameter
                if not isinstance(parameter, ExpressionTreeNode)
                else update_expression_tree(parameter)
                for parameter in node.parameters
            ]

        # this is the main work of this action
        if (node.left and node.left.token_type & LITERAL_TYPE == LITERAL_TYPE) and (
            node.right and node.right.token_type & LITERAL_TYPE == LITERAL_TYPE
        ):
            if node.token_type == NodeType.COMPARISON_OPERATOR:
                value = filter_operations([node.left.value], node.value, [node.right.value])[0]

                return build_literal_node(value)
            if node.token_type == NodeType.BINARY_OPERATOR:
                value = binary_operations([node.left.value], node.value, [node.right.value])[0]
                return build_literal_node(value)
        return node

    # find the in-scope nodes (WHERE AND HAVING)
    selection_nodes = plan.get_nodes_of_type(operators.SelectionNode)

    # killer questions - if any aren't met, bail
    if selection_nodes is None:
        return plan

    # HAVING and WHERE are selection nodes
    for nid in selection_nodes:
        # get the node from the node_id
        operator = plan[nid]
        operator.filter = update_expression_tree(operator.filter)
        if operator.filter is None:
            plan.remove_node(nid, heal=True)

    return plan



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .base.base_blob_storage_adapter import BaseBlobStorageAdapter  # isort: skip
from .base.base_document_storage_adapter import BaseDocumentStorageAdapter  # isort: skip

import pyarrow

from opteryx.config import DATASET_PREFIX_MAPPING
from opteryx.connectors.arrow_connector import ArrowConnector
from opteryx.connectors.aws_s3_connector import AwsS3Connector
from opteryx.connectors.disk_connector import DiskConnector
from opteryx.connectors.gcp_cloudstorage_connector import GcpCloudStorageConnector
from opteryx.connectors.gcp_firestore_connector import GcpFireStoreConnector
from opteryx.connectors.hadro_connector import HadroConnector
from opteryx.connectors.mongodb_connector import MongoDbConnector
from opteryx.connectors.sql_connector import SqlConnector
from opteryx.shared import MaterializedDatasets

WELL_KNOWN_ADAPTERS = {
    "disk": DiskConnector,
    "gcs": GcpCloudStorageConnector,
    "hadro": HadroConnector,
    "firestore": GcpFireStoreConnector,
    "minio": AwsS3Connector,
    "mongodb": MongoDbConnector,
    "s3": AwsS3Connector,
    "sql": SqlConnector,
}

_storage_prefixes = {"information_schema": "InformationSchema"}

if not isinstance(DATASET_PREFIX_MAPPING, dict):  # pragma: no cover
    _storage_prefixes["_"] = "disk"
else:
    for _prefix, _adapter_name in DATASET_PREFIX_MAPPING.items():
        _storage_prefixes[_prefix] = WELL_KNOWN_ADAPTERS.get(
            _adapter_name.lower(), None
        )  # type:ignore


def register_store(prefix, connector, *, remove_prefix: bool = False, **kwargs):
    """add a prefix"""
    if not isinstance(connector, type):  # type: ignore
        # uninstantiated classes aren't a type
        raise ValueError("connectors registered with `register_store` must be uninstantiated.")
    _storage_prefixes[prefix] = {
        "connector": connector,  # type: ignore
        "prefix": prefix,
        "remove_prefix": remove_prefix,
        **kwargs,
    }


def register_df(name, frame):
    """register a pandas or Polars dataframe"""
    # polars (maybe others) - the polars to arrow API is a mess
    if hasattr(frame, "_df"):
        frame = frame._df
    if hasattr(frame, "to_arrow"):
        arrow = frame.to_arrow()
        if not isinstance(arrow, pyarrow.Table):
            arrow = pyarrow.Table.from_batches(arrow)
        register_arrow(name, arrow)
        return
    # pandas
    frame_type = str(type(frame))
    if "pandas" in frame_type:
        register_arrow(name, pyarrow.Table.from_pandas(frame))
        return
    raise ValueError("Unable to register unknown frame type.")


def register_arrow(name, table):
    """register an arrow table"""
    materialized_datasets = MaterializedDatasets()
    materialized_datasets[name] = table
    register_store(name, ArrowConnector)


def connector_factory(dataset):
    prefix = dataset.split(".")[0]
    connector_entry: dict = {}
    if prefix in _storage_prefixes:
        connector_entry = _storage_prefixes[prefix].copy()  # type: ignore
        connector = connector_entry.pop("connector")
    else:
        connector = _storage_prefixes.get("_")

    prefix = connector_entry.pop("prefix", "")
    remove_prefix = connector_entry.pop("remove_prefix", False)

    return connector(prefix=prefix, remove_prefix=remove_prefix, **connector_entry)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This implements an in-memory cache.

We use an LRU-K(2) to maintain the size of the cache, this is a variation of the naive
LRU algorithm.

We're using a dictionary and moving items to the top of the dictionary when it's
accessed. This relies on Python dictionaries being ordered.
"""

import io
from typing import Iterable

from opteryx.managers.kvstores import BaseKeyValueStore
from opteryx.utils.lru_2 import LRU2


class MemoryCache(BaseKeyValueStore):
    slots = ("_lru2",)

    def __init__(self, **kwargs):
        """
        Parameters:
            size: int (optional)
                The maximim number of items maintained in the cache.
        """
        size = int(kwargs.get("size", 50))
        self._lru2 = LRU2(size=size)

    def get(self, key):
        value = self._lru2.get(key)
        if value:
            return io.BytesIO(value)

    def set(self, key, value):
        value.seek(0, 0)
        ret = self._lru2.set(key, value.read())
        value.seek(0, 0)
        return ret

    def contains(self, keys: Iterable) -> Iterable:
        return list(set(keys).intersection(self._lru2.keys))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Projection Node

This is a SQL Query Execution Plan Node.

This Node eliminates columns that are not needed in a Relation. This is also the Node
that performs column renames.
"""
import time
from typing import Iterable

import pyarrow

from opteryx.exceptions import SqlError
from opteryx.managers.expression import LITERAL_TYPE
from opteryx.managers.expression import ExpressionTreeNode
from opteryx.managers.expression import NodeType
from opteryx.managers.expression import evaluate_and_append
from opteryx.managers.expression import format_expression
from opteryx.models import QueryProperties
from opteryx.operators import BasePlanNode
from opteryx.utils import random_int


class ProjectionNode(BasePlanNode):
    def __init__(self, properties: QueryProperties, **config):
        """
        Attribute Projection, remove unwanted columns and performs column renames.
        """
        super().__init__(properties=properties)
        self._projection: dict = {}
        self._expressions = []

        projection = config.get("projection")
        for attribute in projection:
            while attribute.token_type == NodeType.NESTED:
                attribute = attribute.centre
            if attribute.token_type == NodeType.WILDCARD and attribute.value is not None:
                # qualified wildcard, e.g. table.*
                self._projection[attribute.value] = None
            elif attribute.token_type in (
                NodeType.FUNCTION,
                NodeType.AGGREGATOR,
                NodeType.COMPLEX_AGGREGATOR,
                NodeType.BINARY_OPERATOR,
                NodeType.COMPARISON_OPERATOR,
            ) or (attribute.token_type & LITERAL_TYPE == LITERAL_TYPE):
                new_column_name = format_expression(attribute)
                self._projection[new_column_name] = attribute.alias
                self._expressions.append(attribute)
            elif attribute.token_type == NodeType.IDENTIFIER:
                self._projection[attribute.value] = attribute.alias
            elif isinstance(attribute, ExpressionTreeNode):
                new_column_name = format_expression(attribute)
                self._projection[new_column_name] = attribute.alias
                self._expressions.append(attribute)
            else:
                self._projection[attribute] = None

    @property
    def config(self):  # pragma: no cover
        return str(self._projection)

    @property
    def name(self):  # pragma: no cover
        return "Projection"

    def execute(self) -> Iterable:
        if len(self._producers) != 1:  # pragma: no cover
            raise SqlError(f"{self.name} on expects a single producer")

        morsels = self._producers[0]  # type:ignore

        # if we have nothing to do, move along
        if self._projection == {"*": None}:
            yield from morsels.execute()
            return

        # we can't do much with this until we have a chunk to read the metadata from
        columns = None
        # we want to avoid collisions in internal column names
        seed = str(random_int() + 1)

        for morsel in morsels.execute():
            # If any of the columns are FUNCTIONs, we need to evaluate them
            start_time = time.time_ns()
            _columns, _, morsel = evaluate_and_append(self._expressions, morsel, seed)
            self.statistics.time_evaluating += time.time_ns() - start_time

            # first time round we're going work out what we need from the metadata
            if columns is None:
                projection = []
                columns = _columns
                for key in self._projection:
                    if isinstance(key, tuple):
                        relation = key[0]
                        projection.extend(columns.get_columns_from_source(relation))
                    else:
                        projection.append(columns.get_column_from_alias(key, only_one=True))

            morsel = morsel.select(projection)  # type:ignore

            if len(projection) != len(set(projection)):
                raise SqlError(
                    "SELECT statement contains multiple references to the same column, perhaps as aliases or with qualifiers."
                )

            # then we rename the attributes
            if any([v is not None for k, v in self._projection.items()]):  # type:ignore
                existing_columns = morsel.column_names
                for k, v in self._projection.items():
                    if isinstance(v, list) and len(v) != 0:
                        v = v[0]
                    if v and v not in existing_columns:
                        column_name = columns.get_column_from_alias(k, only_one=True)
                        columns.set_preferred_name(column_name, v)
                morsel = columns.apply(morsel)

            yield morsel



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Write to a log with a fixed number of entries, as the log fills the older entries are removed
from the log file.
"""
import os


class RollingLog:
    _instance = None

    def __new__(cls, log_file: str, max_entries: int = 100, block_size: int = 1024 * 1024):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance.log_file = log_file  # type:ignore
            cls._instance.max_entries = max_entries  # type:ignore
            cls._instance.block_size = block_size  # type:ignore
            if not os.path.exists(log_file):
                open(log_file, "wb")
        return cls._instance

    def append(self, entry):
        # open the log file in binary mode
        with open(self.log_file, "a", encoding="UTF8") as log_file:  # type:ignore
            log_file.write(entry + "\n")
        if sum(1 for _ in self.scan()) > self.max_entries:  # type:ignore
            with open(self.log_file, "r+b") as f:  # type:ignore
                while f.read(1) != b"\n":
                    pass
                remaining_data = f.read()
                f.seek(0)
                f.write(remaining_data)
                f.truncate()

    def scan(self):
        # open the log file in binary mode
        with open(self.log_file, "r", encoding="UTF8") as log_file:  # type:ignore
            # read the current position in the circular buffer
            chunk = log_file.read(self.block_size)  # type:ignore
            # keep reading until there's no more data left
            while chunk:
                # split the chunk by newlines
                lines = chunk.split("\n")
                # yield each line
                for line in lines:
                    if line:
                        yield line
                # read the next chunk
                chunk = log_file.read(self.block_size)  # type:ignore



"""
Original code modified for Opteryx.
"""
from enum import Enum
from ipaddress import IPv4Address
from ipaddress import IPv4Network

import numpy
import pyarrow
from pyarrow import compute

from opteryx.constants.attribute_types import OPTERYX_TYPES
from opteryx.constants.attribute_types import PARQUET_TYPES

from .helpers import columns_to_array

# Added for Opteryx, comparisons in filter_operators updated to match
# this set is from sqloxide
FILTER_OPERATORS = {
    "Eq",
    "NotEq",
    "Gt",
    "GtEq",
    "Lt",
    "LtEq",
    "Like",
    "ILike",
    "NotLike",
    "NotILike",
    "InList",
    "SimilarTo",
    "NotSimilarTo",
    "PGRegexMatch",
    "NotPGRegexMatch",
    "PGRegexNotMatch",
    "PGRegexIMatch",  # "~*"
    "NotPGRegexIMatch",  # "!~*"
    "PGRegexNotIMatch",  # "!~*"
    "BitwiseOr",  # |
}


class TOKEN_TYPES(str, Enum):
    BOOLEAN = "BOOLEAN"
    NUMERIC = "NUMERIC"
    LIST = "LIST"
    VARCHAR = "VARCHAR"
    STRUCT = "STRUCT"
    TIMESTAMP = "TIMESTAMP"
    OTHER = "OTHER"
    IDENTIFIER = "IDENTIFIER"
    WILDCARD = "WILDCARD"
    QUERY_PLAN = "QUERY_PLAN"
    FUNCTION = "FUNCTION"
    INTERVAL = "INTERVAL"


PYTHON_TYPES = {
    "bool": OPTERYX_TYPES.BOOLEAN,
    "datetime": OPTERYX_TYPES.TIMESTAMP,
    "date": OPTERYX_TYPES.TIMESTAMP,
    "dict": OPTERYX_TYPES.STRUCT,
    "int": OPTERYX_TYPES.NUMERIC,  # INTEGER
    "float": OPTERYX_TYPES.NUMERIC,  # FLOAT
    "float64": OPTERYX_TYPES.NUMERIC,  # IS THIS USED?
    "Decimal": OPTERYX_TYPES.NUMERIC,  # DECIMAL
    "str": OPTERYX_TYPES.VARCHAR,
    "tuple": OPTERYX_TYPES.LIST,
    "list": OPTERYX_TYPES.LIST,
    "set": OPTERYX_TYPES.LIST,
    # INTERVAL?
}


def _get_type(var):
    # added for Opteryx
    if isinstance(var, (numpy.ndarray)):
        _type = str(var.dtype)
        if _type.startswith("<U"):
            _type = "string"
        return PARQUET_TYPES.get(_type, f"UNSUPPORTED ({str(var.dtype)})")
    if isinstance(var, (pyarrow.Array)):
        return PARQUET_TYPES.get(str(var.type), f"UNSUPPORTED ({str(var.type)})")
    if isinstance(var, list):
        return PYTHON_TYPES.get(type(var[0]).__name__, f"UNSUPPORTED ({type(var[0]).__name__})")
    type_name = type(var).__name__  # pragma: no cover
    return PYTHON_TYPES.get(type_name, f"OTHER ({type_name})")  # pragma: no cover


def _check_type(operation, provided_type, valid_types):
    # added for Opteryx
    if provided_type not in valid_types:  # pragma: no cover
        raise TypeError(
            f"Cannot use the {operation} operation on a {provided_type} column, a {valid_types} column is required."
        )


def filter_operations_for_display(arr, operator, value):
    """
    Wrapped for Opteryx added to correctly handle null semantics.

    This returns an array with tri-state boolean (tue/false/none). This is
    for use where an expression result is deplayed to users.
    """

    # if the input is a table, get the first column
    if isinstance(value, pyarrow.Table):  # pragma: no cover
        value = [value.columns[0].to_numpy()]

    # work out which rows we're going to actually evaluate
    # we're working out if either array has a null value so we can exclude them
    # from the actual evaluation.
    #   True = values, False = null
    record_count = len(arr)
    null_arr = compute.is_null(arr, nan_is_null=True)
    null_val = compute.is_null(value, nan_is_null=True)
    null_positions = numpy.logical_or(null_arr, null_val)

    # if there's no non-null values, stop here
    if all(null_positions):
        return numpy.full(record_count, None)

    any_null = any(null_positions)
    null_positions = numpy.invert(null_positions)

    compressed = False
    if any_null and isinstance(arr, numpy.ndarray) and isinstance(value, numpy.ndarray):
        # if we have nulls and both columns are numpy arrays, we can speed things
        # up by removing the nulls from the calculations, we add the rows back in
        # later
        arr = arr.compress(null_positions)
        value = value.compress(null_positions)
        compressed = True

    # do the evaluation
    results_mask = _inner_filter_operations(arr, operator, value)

    if compressed:
        # fill the result set
        results = numpy.full(record_count, -1, numpy.int8)
        results[numpy.nonzero(null_positions)] = results_mask
        # build tri-state response
        return [bool(r) if r != -1 else None for r in results]

    return results_mask


def filter_operations(arr, operator, value):
    """
    Wrapped for Opteryx added to correctly handle null semantics.

    This is used where the filter actually filters records so is bi-state
    (true/false) where null is coaleced to false.
    """

    # if the input is a table, get the first column
    if isinstance(value, pyarrow.Table):  # pragma: no cover
        value = [value.columns[0].to_numpy()]

    # work out which rows we're going to actually evaluate
    # we're working out if either array has a null value so we can exclude them
    # from the actual evaluation.
    #   True = values, False = null
    record_count = len(arr)
    null_arr = compute.is_null(arr, nan_is_null=True)
    null_val = compute.is_null(value, nan_is_null=True)
    null_positions = numpy.logical_or(null_arr, null_val)

    # if there's no non-null values, stop here
    if all(null_positions):
        return numpy.full(record_count, False)

    any_nulls = any(null_positions)
    null_positions = numpy.invert(null_positions)

    compressed = False
    if any_nulls and isinstance(arr, numpy.ndarray) and isinstance(value, numpy.ndarray):
        # if we have nulls and both columns are numpy arrays, we can speed things
        # up by removing the nulls from the calculations, we add the rows back in
        # later
        arr = arr.compress(null_positions)
        value = value.compress(null_positions)
        compressed = True

    # do the evaluation
    results_mask = _inner_filter_operations(arr, operator, value)

    # fill the result set
    if compressed:
        results = numpy.full(record_count, False, numpy.bool_)
        results[numpy.nonzero(null_positions)] = results_mask
        return results

    return results_mask


# Filter functionality
def _inner_filter_operations(arr, operator, value):
    """
    Execute filter operations, this returns an array of the indexes of the rows that
    match the filter
    """
    # ADDED FOR OPTERYX
    identifier_type = _get_type(arr)
    literal_type = _get_type(value)

    if operator == "Eq":
        if identifier_type != literal_type and value is not None:  # pragma: no cover
            raise TypeError(
                f"Type mismatch, unable to compare {identifier_type} with {literal_type}"
            )
        return compute.equal(arr, value).to_numpy(False).astype(dtype=bool)
    elif operator == "NotEq":
        return compute.not_equal(arr, value).to_numpy(False).astype(dtype=bool)
    elif operator == "Lt":
        return compute.less(arr, value).to_numpy(False).astype(dtype=bool)
    elif operator == "Gt":
        return compute.greater(arr, value).to_numpy(False).astype(dtype=bool)
    elif operator == "LtEq":
        return compute.less_equal(arr, value).to_numpy(False).astype(dtype=bool)
    elif operator == "GtEq":
        return compute.greater_equal(arr, value).to_numpy(False).astype(dtype=bool)
    elif operator == "InList":
        # MODIFIED FOR OPTERYX
        # some of the lists are saved as sets, which are faster than searching numpy
        # arrays, even with numpy's native functionality - choosing the right algo
        # is almost always faster than choosing a fast language.
        return numpy.array([a in value[0] for a in arr], dtype=numpy.bool_)  # [#325]?
    elif operator == "NotInList":
        # MODIFIED FOR OPTERYX - see comment above
        return numpy.array([a not in value[0] for a in arr], dtype=numpy.bool_)  # [#325]?
    elif operator == "Contains":
        # ADDED FOR OPTERYX
        return numpy.array([None if v is None else (arr[0] in v) for v in value], dtype=numpy.bool_)
    elif operator == "NotContains":
        # ADDED FOR OPTERYX
        return numpy.array(
            [None if v is None else (arr[0] not in v) for v in value], dtype=numpy.bool_
        )  # [#325]?
    elif operator == "Like":
        # MODIFIED FOR OPTERYX
        # null input emits null output, which should be false/0
        _check_type("LIKE", identifier_type, (TOKEN_TYPES.VARCHAR))
        return compute.match_like(arr, value[0]).to_numpy(False).astype(dtype=bool)  # [#325]
    elif operator == "NotLike":
        # MODIFIED FOR OPTERYX - see comment above
        _check_type("NOT LIKE", identifier_type, (TOKEN_TYPES.VARCHAR))
        matches = compute.match_like(arr, value[0]).to_numpy(False).astype(dtype=bool)  # [#325]
        return numpy.invert(matches)
    elif operator == "ILike":
        # MODIFIED FOR OPTERYX - see comment above
        _check_type("ILIKE", identifier_type, (TOKEN_TYPES.VARCHAR))
        return (
            compute.match_like(arr, value[0], ignore_case=True).to_numpy(False).astype(dtype=bool)
        )  # [#325]
    elif operator == "NotILike":
        # MODIFIED FOR OPTERYX - see comment above
        _check_type("NOT ILIKE", identifier_type, (TOKEN_TYPES.VARCHAR))
        matches = compute.match_like(arr, value[0], ignore_case=True)  # [#325]
        return numpy.invert(matches)
    elif operator in ("PGRegexMatch", "SimilarTo"):
        # MODIFIED FOR OPTERYX - see comment above
        _check_type("~", identifier_type, (TOKEN_TYPES.VARCHAR))
        return (
            compute.match_substring_regex(arr, value[0]).to_numpy(False).astype(dtype=bool)
        )  # [#325]
    elif operator in ("PGRegexNotMatch", "NotSimilarTo"):
        # MODIFIED FOR OPTERYX - see comment above
        _check_type("!~", identifier_type, (TOKEN_TYPES.VARCHAR))
        matches = compute.match_substring_regex(arr, value[0])  # [#325]
        return numpy.invert(matches)
    elif operator == "PGRegexIMatch":
        # MODIFIED FOR OPTERYX - see comment above
        _check_type("~*", identifier_type, (TOKEN_TYPES.VARCHAR))
        return (
            compute.match_substring_regex(arr, value[0], ignore_case=True)
            .to_numpy(False)
            .astype(dtype=bool)
        )  # [#325]
    elif operator == "PGRegexNotIMatch":
        # MODIFIED FOR OPTERYX - see comment above
        _check_type("!~*", identifier_type, (TOKEN_TYPES.VARCHAR))
        matches = compute.match_substring_regex(arr, value[0], ignore_case=True)  # [#325]
        return numpy.invert(matches)
    elif operator == "BitwiseOr":
        try:
            # parse right to be a list of IPs
            value = IPv4Network(value[0], strict=False)
            # is left in right
            result = []
            for address in arr:
                if address:
                    result.append(IPv4Address(address) in value)
                else:
                    result.append(None)
            return result
        except Exception as e:
            print(e)
            raise NotImplementedError("`|` can only be used to test IP address containment.")
    else:
        raise NotImplementedError(f"Operator {operator} is not implemented!")  # pragma: no cover


# Drop duplicates
def drop_duplicates(table, columns=None):
    """
    drops duplicates, keeps the first of the set

    MODIFIED FOR OPTERYX
    """
    # Gather columns to arr
    arr = columns_to_array(table, (columns if columns else table.column_names))
    values, indices = numpy.unique(arr, return_index=True)
    del values
    return table.take(indices)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Optimization Rule - Split Conjunctive Predicates (ANDs)

Type: Heuristic
Goal: Reduce rows
"""
from opteryx import operators
from opteryx.managers.expression import NodeType
from opteryx.utils import random_string


def split_conjunctive_predicates(plan, properties):
    """
    Conjunctive Predicates (ANDs) can be split and executed in any order to get the
    same result. This means we can split them into separate steps in the plan.

    The reason for splitting is two-fold:

    1)  Smaller expressions are easier to move around the query plan as they have fewer
        dependencies.
    2)  Executing predicates like this means each runs in turn, filtering out some of
        the records meaning susequent predicates will be operating on fewer records,
        which is generally faster. We can also order these predicates to get a faster
        result, balancing the selectivity (get rid of more records faster) vs cost of
        the check (a numeric check is faster than a string check)
    """

    def _inner_split(plan, nid, operator):
        selection = operator.filter
        if selection.token_type != NodeType.AND:
            return plan

        # get the left and right filters
        left_node = operators.SelectionNode(filter=selection.left, properties=properties)
        right_node = operators.SelectionNode(filter=selection.right, properties=properties)
        # insert them into the plan and remove the old node
        # we're chaining the new operators
        uid = random_string()  # avoid collisions
        plan.insert_node_before(f"{nid}-{uid}-right", right_node, nid)
        plan.insert_node_before(f"{nid}-{uid}-left", left_node, f"{nid}-{uid}-right")
        plan.remove_node(nid, heal=True)

        # recurse until we get to a non-AND condition
        plan = _inner_split(plan, f"{nid}-{uid}-right", right_node)
        plan = _inner_split(plan, f"{nid}-{uid}-left", left_node)

        return plan

    # find the in-scope nodes
    selection_nodes = plan.get_nodes_of_type(operators.SelectionNode)

    # killer questions - if any aren't met, bail
    if selection_nodes is None:
        return plan

    # HAVING and WHERE are selection nodes
    for nid in selection_nodes:
        # get the node from the node_id
        operator = plan[nid]
        plan = _inner_split(plan, nid, operator)

    return plan



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Optimization Rule - Predicate Pushdown

Type: Heuristic
Goal: Reduce rows
"""
from opteryx import config
from opteryx import operators
from opteryx.managers.expression import NodeType


def predicate_pushdown(plan, properties):
    """
    Initial implementation of Predicate Pushdown, this puts selections (filters) as
    close to the read as possible, including offloading them to external services.

    This has the benefit that we have less data to pass around the system, and
    where we can offload to other systems (e.g. FireStore) we don't need to
    transport that data over the network.

    We do this work by removing the selection node from the query plan and adding
    it to the filter in the reader node.

    This initial implementation has a lot of limitations:

    - only a single reader can exist in the plan
    - only some readers support pushdown
    - no GROUP BY / Aggregate can exist in the plan

    """
    # find the in-scope nodes
    selection_nodes = plan.get_nodes_of_type(operators.SelectionNode)
    reader_nodes = plan.get_nodes_of_type(
        (
            operators.BlobReaderNode,
            operators.CollectionReaderNode,
            operators.FileReaderNode,
            operators.SqlReaderNode,
        )
    )

    # killer questions - if any aren't met, bail
    if selection_nodes is None:
        return plan
    if len(plan.get_nodes_of_type(operators.AggregateNode)) > 0:
        # don't try to work out if it's a HAVING or GROUP BY
        # a HAVING without a GROUP BY should be okay
        return plan
    if len(reader_nodes) != 1:
        # don't try to work out which reader to push to
        return plan
    reader_node = plan[reader_nodes[0]]
    if not reader_node.can_push_selection:
        # not all readers support pushdown
        return plan

    # WHERE are selection nodes
    for nid in selection_nodes:
        # get the node from the node_id
        operator = plan[nid]
        # only add simple predicates (makes ANDs)
        if operator.filter.token_type == NodeType.COMPARISON_OPERATOR:
            if config.ONLY_PUSH_EQUALS_PREDICATES and operator.filter.value != "Eq":
                continue
            if reader_node.push_predicate(operator.filter):
                plan.remove_node(nid, heal=True)

    return plan



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
SQL Reader Node

This is a SQL Query Execution Plan Node.

This Node primarily is used for reading NoSQL sources like MongoDB and Firestore.
"""
import time
from typing import Iterable

import pyarrow

from opteryx.managers.expression import NodeType
from opteryx.models import QueryProperties
from opteryx.models.columns import Columns
from opteryx.operators import BasePlanNode


class SqlReaderNode(BasePlanNode):
    def __init__(self, properties: QueryProperties, **config):
        """
        The Collection Reader Node is responsible for reading the relevant documents
        from a NoSQL document store and returning a Table/Relation.
        """
        super().__init__(properties=properties)

        self._alias = config.get("alias")
        self._dataset = config["dataset"]
        self._reader = config.get("reader")

        # pushed down selection/filter
        self._selection = config.get("selection")

        self._disable_selections = "NO_PUSH_SELECTION" in config.get("hints", [])

    @property
    def config(self):  # pragma: no cover
        if self._alias:
            return f"{self._dataset} => {self._alias}"
        return f"{self._dataset}"

    @property
    def name(self):  # pragma: no cover
        return "SQL Reader"

    @property
    def can_push_selection(self):
        return not self._disable_selections

    def push_predicate(self, predicate):
        # SQL reader currently assumes the right parameter is a literal
        if predicate.left.token_type == predicate.right.token_type == NodeType.IDENTIFIER:
            return False
        if self.can_push_selection:
            return self._reader.push_predicate(predicate)
        return False

    def execute(self) -> Iterable:
        metadata = None

        for morsel in self._reader.read_records(dataset=self._dataset):
            start_read = time.time_ns()
            if not isinstance(morsel, pyarrow.Table):
                morsel = pyarrow.Table.from_pylist(morsel)

            self.statistics.time_data_read += time.time_ns() - start_read
            self.statistics.rows_read += morsel.num_rows
            self.statistics.bytes_processed_data += morsel.nbytes
            self.statistics.document_chunks += 1

            if metadata is None:
                morsel = Columns.create_table_metadata(
                    table=morsel,
                    expected_rows=None,  # TODO: can we get this?
                    name=self._dataset,
                    table_aliases=[self._alias],
                    disposition="collection",
                    path=self._dataset,
                )
                metadata = Columns(morsel)
                self.statistics.tables_read += 1
                self.statistics.columns_read += len(morsel.column_names)
            else:
                morsel = metadata.apply(morsel)

            yield morsel



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing

from opteryx import config
from opteryx.connectors import BaseDocumentStorageAdapter


class HadroConnector(BaseDocumentStorageAdapter):
    def __init__(self, *args, prefix: str = "", remove_prefix: bool = False, **kwargs):
        BaseDocumentStorageAdapter.__init__(
            self, *args, prefix=prefix, remove_prefix=remove_prefix, **kwargs  # type: ignore
        )
        self._remove_prefix = remove_prefix
        self._prefix = prefix

    def get_document_count(self, collection) -> int:  # pragma: no cover
        """
        Return an interable of blobs/files

        FireStore currently doesn't support this
        """
        return -1

    def read_documents(self, collection, morsel_size: typing.Union[int, None] = None):
        """
        Return a morsel of documents
        """
        from hadrodb import HadroDB

        if morsel_size is None:
            morsel_size = config.MORSEL_SIZE

        queried_collection = collection
        if self._remove_prefix:
            if collection.startswith(f"{self._prefix}."):
                queried_collection = collection[len(self._prefix) + 1 :]
        queried_collection = queried_collection.replace(".", "/")

        hadro = HadroDB(collection=queried_collection)

        for morsel in self.chunk_dictset((hadro[doc] for doc in hadro.keys()), morsel_size):
            yield morsel



import functools

import numpy

APOLLO_11_DURATION: int = 703115  # we need a constant to use as a seed


def groupify_array(arr):
    # Input: Pyarrow/Numpy array
    # Output:
    #   - 1. Unique values
    #   - 2. Count per unique
    #   - 3. Sort index
    #   - 4. Begin index per unique

    # UPDATED FOR OPTERYX
    dic, counts = numpy.unique(arr, return_counts=True, equal_nan=True)
    sort_idx = numpy.argsort(arr)
    return dic, counts, sort_idx, [0] + numpy.cumsum(counts)[:-1].tolist()


def _hash_value(val, nan=numpy.nan):
    # Added for Opteryx - Original code had bugs relating to distinct and nulls
    if isinstance(val, dict):
        return _hash_value(tuple(val.values()))
    if isinstance(val, (list, numpy.ndarray, tuple)):
        # XOR is faster however, x ^ x == y ^ y but x != y, so we don't use it
        return functools.reduce(lambda x, y: _hash_value(f"{y}:{x}", 0), val, APOLLO_11_DURATION)
    if val != val or val is None:  # nosemgrep
        # nan is a float, but hash is an int, sometimes we need this to be an int
        return nan
    return hash(val)


def columns_to_array(table, columns):
    """modified for Opteryx"""
    # used for distinct
    columns = [columns] if isinstance(columns, str) else list(dict.fromkeys(columns))

    if len(columns) == 1:
        # FIX https://github.com/mabel-dev/opteryx/issues/98
        # hashing NULL doesn't result in the same value each time
        # FIX https://github.com/mabel-dev/opteryx/issues/285
        # null isn't able to be sorted - replace with nan
        column_values = table.column(columns[0]).to_numpy()
        return numpy.array([_hash_value(el) for el in column_values])

    columns = sorted(set(table.column_names).intersection(columns))
    values = (c.to_numpy() for c in table.select(columns).itercolumns())
    return numpy.array([_hash_value(x) for x in zip(*values)])


def columns_to_array_denulled(table, columns):
    """added for Opteryx"""
    # Used for joins - nulls don't join so we can remove them
    # The columns need to be read in the order provided, so don't set() or order()
    # them anywhere here.
    columns = [columns] if isinstance(columns, str) else columns
    if len(columns) == 1:
        column_values = table.column(columns[0]).drop_null().to_numpy()
        return numpy.array([_hash_value(el) for el in column_values])

    values = (c.to_numpy() for c in table.select(columns).drop_null().itercolumns())
    return numpy.array([_hash_value(x) for x in zip(*values)])



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
LRU-K evicts the morsel whose K-th most recent access is furthest in the past.

This is a basic implementation of LRU-2, which evicts entries according to the time of their
penultimate access. The main benefit of this approach is to prevent a problem when the items being
getted exceeds the number of items in the cache. A classic LRU will evict and repopulate the cache
for every call. LRU-2 reduces the likelihood of this, but not preferring the MRU item to be
retained.

LRU-K should be used in conjunction with eviction limits per query - this appears to broadly be
the solution used by Postgres. This can be supported by the calling function using the return from
the .set call to determine if an item was evicted from the cache.

This can also be used as the index for an external cache (for example in plasma), where the set()
returns the evicted item which the calling function can then evict from the external cache.
"""

import time
from collections import defaultdict


class LRU2:
    def __init__(self, k=2, size=50):
        self.k = k
        self.size = size
        self.cache = {}
        self.access_history = defaultdict(list)
        self.hits = 0
        self.misses = 0
        self.evictions = 0

    def get(self, key):
        # Check if the key is in the cache
        if key in self.cache:
            # Update the access history for the key
            self.access_history[key].append(time.monotonic_ns())
            if len(self.access_history[key]) > self.k:
                self.access_history[key].pop(0)
            # Increment the hit count and return the value
            self.hits += 1
            return self.cache[key]
        else:
            # Increment the miss count and return None
            self.misses += 1
            return None

    def set(self, key, value):
        # Check if the key is already in the cache
        if key in self.cache:
            return None

        # Add the new key-value pair to the cache
        self.cache[key] = value

        # Update the access history for the key
        self.access_history[key].append(time.monotonic_ns())

        # If the cache is full, evict the least recently used item
        if len(self.cache) > self.size:
            # Find the key with the oldest access time and remove it from the cache
            oldest_key = None
            oldest_access_time = float("inf")
            for k in self.cache:
                if self.access_history[k][0] < oldest_access_time:
                    oldest_key = k
                    oldest_access_time = self.access_history[k][0]
            self.cache.pop(oldest_key)
            self.access_history.pop(oldest_key)
            self.evictions += 1
            return oldest_key

        return None

    @property
    def keys(self):  # pragma: no-cover
        return list(self.cache.keys())

    @property
    def stats(self):
        # return hits, misses, evictions
        return (self.hits, self.misses, self.evictions)

    def reset(self, reset_stats: bool = False):
        self.cache = {}
        if reset_stats:
            self.hits = 0
            self.misses = 0
            self.evictions = 0



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Optimization Rule - Demorgan's Laws

Type: Heuristic
Goal: Preposition for following actions
"""
from opteryx import operators
from opteryx.managers.expression import ExpressionTreeNode
from opteryx.managers.expression import NodeType


def apply_demorgans_law(plan, properties):
    """
    This action aims to create more opportunity for other rules to act on. Demorgan's
    Laws allow the conversion of ORs to ANDs through Negations.

    By converting ORs to ANDs the 'split conjuctive predicates' action will have more
    to act on. By changing how the Negations are expressed, the 'eliminate negations'
    action will have more to act on.

    The core of this action is taking advantage of the following:

        not (A or B) = (not A) and (not B)
    """

    def update_expression_tree(node):
        """
        Walk a expression tree collecting all the nodes of a specified type.
        """
        # this is the main work of this action
        if node.token_type == NodeType.NESTED:
            return update_expression_tree(node.centre)
        if node.token_type == NodeType.NOT:
            centre_node = node.centre

            # break out of nesting
            if centre_node.token_type == NodeType.NESTED:
                centre_node = centre_node.centre

            # do we have a NOT (a or b)?
            if centre_node.token_type == NodeType.OR:
                # rewrite to (not A) and (not B)
                a_side = ExpressionTreeNode(
                    NodeType.NOT, centre=update_expression_tree(centre_node.left)
                )
                b_side = ExpressionTreeNode(
                    NodeType.NOT, centre=update_expression_tree(centre_node.right)
                )
                return ExpressionTreeNode(NodeType.AND, left=a_side, right=b_side)

        # below here is generic to walk the tree
        node.left = None if node.left is None else update_expression_tree(node.left)
        node.centre = None if node.centre is None else update_expression_tree(node.centre)
        node.right = None if node.right is None else update_expression_tree(node.right)
        if node.parameters:
            node.parameters = [
                parameter
                if not isinstance(parameter, ExpressionTreeNode)
                else update_expression_tree(parameter)
                for parameter in node.parameters
            ]
        return node

    # find the in-scope nodes (WHERE AND HAVING)
    selection_nodes = plan.get_nodes_of_type(operators.SelectionNode)

    # killer questions - if any aren't met, bail
    if selection_nodes is None:
        return plan

    # HAVING and WHERE are selection nodes
    for nid in selection_nodes:
        # get the node from the node_id
        operator = plan[nid]
        operator.filter = update_expression_tree(operator.filter)

    return plan



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from opteryx.exceptions import NotSupportedError
from opteryx.managers.expression import NodeType

PUSHABLE_OPERATORS = {
    # these are almost universally supported
    "Gt": ">",
    "Lt": "<",
    "Eq": "==",  # usually ==, sometimes =
    "NotEq": "!=",  # usually !=, sometimes <>
    "GtEq": ">=",
    "LtEq": "<=",
    "Like": "Like",
}


def to_dnf(root):
    """
    Convert a filter to the form used by the selection pushdown

    Version 1 only does simple predicate filters in the form
        (identifier, operator, literal)
    although we can form conjuntions (ANDs) by chaining them.

    Return None if we can't convert, or don't support the filter.
    """

    def _predicate_to_dnf(root):
        # Reduce look-ahead effort by using Exceptions to control flow
        if root.token_type == NodeType.AND:
            left = _predicate_to_dnf(root.left)
            right = _predicate_to_dnf(root.right)
            if not isinstance(left, list):
                left = [left]
            if not isinstance(right, list):
                right = [right]
            left.extend(right)
            return left
        if root.token_type != NodeType.COMPARISON_OPERATOR:
            raise NotSupportedError()
        if not root.value in PUSHABLE_OPERATORS:
            # not all operators are universally supported
            raise NotSupportedError()
        if root.left.token_type != NodeType.IDENTIFIER:
            raise NotSupportedError()
        if root.left.token_type in (
            NodeType.LITERAL_NUMERIC,
            NodeType.LITERAL_VARCHAR,
        ):
            # not all operands are universally supported
            raise NotSupportedError()
        return (root.left.value, PUSHABLE_OPERATORS[root.value], root.right.value)

    try:
        dnf = _predicate_to_dnf(root)
        if not isinstance(dnf, list):
            dnf = [dnf]
    except NotSupportedError:
        return None
    return dnf


class PredicatePushable:
    __slots__ = ("_predicates", "supported_ops")

    def __init__(self, *args, **kwargs):
        self._predicates = []
        self.supported_ops = list(PUSHABLE_OPERATORS.values())

    def push_predicate(self, predicate):
        """
        Push the predicate to the set - this creates a set of ANDs
        """
        dnfed = to_dnf(predicate)
        if dnfed is None:
            # we can't push all predicates everywhere
            return False
        if not all(d[1] in self.supported_ops for d in dnfed):
            return False
        self._predicates.extend(dnfed)
        return True



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Base Inner Reader for blob and file stores
"""
import abc
import datetime
from typing import Iterable
from typing import List
from typing import Union

from opteryx.utils import dates
from opteryx.utils import paths


class BaseBlobStorageAdapter(abc.ABC):
    __mode__ = "Blob"

    def __init__(self, prefix: str, remove_prefix: bool = False) -> None:
        self._prefix = prefix
        self._remove_prefix = remove_prefix

    def get_partitions(
        self,
        *,
        dataset: str,
        partitioning: Iterable = None,  # ("year_{yyyy}", "month_{mm}", "day_{dd}"),
        start_date: Union[datetime.datetime, datetime.date, str] = None,
        end_date: Union[datetime.datetime, datetime.date, str] = None,
    ) -> List:
        """
        Get partitions doesn't confirm the partitions exist, it just creates a list
        of candidate partitions.
        """
        import pathlib

        # apply the partitioning to the dataset name
        if not dataset.endswith("/"):  # pragma: no cover
            dataset += "/"
        if not isinstance(partitioning, (list, set, tuple)):
            partitioning = [partitioning]
        if partitioning:
            dataset += "/".join(partitioning) + "/"

        # make sure we're dealing with dates
        start_date = dates.parse_iso(start_date) or datetime.datetime.utcnow()
        end_date = dates.parse_iso(end_date) or datetime.datetime.utcnow()

        partitions = []
        # we're going to iterate over the date range and get the name of the partition for
        # this dataset on this data - without knowing if the partition exists
        for delta in range(int((end_date - start_date).days) + 1):
            working_date = start_date + datetime.timedelta(delta)
            partitions.append(pathlib.Path(paths.build_path(path=dataset, date=working_date)))

        return partitions

    def get_blob_list(self, partition=None) -> Iterable:  # pragma: no cover
        """
        Return an interable of blobs/files
        """
        raise NotImplementedError("get_blob_list not implemented")

    def read_blob(self, blob_name: str) -> bytes:  # pragma: no cover
        """
        Return a filelike object
        """
        raise NotImplementedError("read_blob not implemented")

    def write_blob(self, name, content):  # pragma: no cover
        raise NotImplementedError("write_blob not implemented")

    def remove_blob(self, name):  # pragma: no cover
        raise NotImplementedError("remove_blob not implemented")



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The Execution Tree is the Graph which defines a Query Plan.

The execution tree contains functionality to:

- build and define the plan
- execute the plan
- manipulate the plan

"""

import pyarrow

from opteryx.exceptions import DatabaseError
from opteryx.third_party.travers import Graph


class ExecutionTree(Graph):
    """
    The execution tree is defined separately to the planner to simplify the
    complex code which is the planner from the tree that describes the plan.
    """

    def get_nodes_of_type(self, operator):
        """
        Traverse the plan looking for nodes of a given type or set of types.

        Parameters:
            operator: BaseOperator or Iterable of BaseOperator
                The operator type(s) to match.
        Yields:
            string
                nids of matching nodes.
        """

        def _inner(operator):
            if not isinstance(operator, (list, set)):
                operator = tuple([operator])
            if not isinstance(operator, tuple):
                operator = tuple(operator)
            for nid, item in list(self._nodes.items()):
                if isinstance(item, operator):
                    yield nid

        return list(_inner(operator))

    def execute(self):
        """
        This implements a 'pull' model execution engine. It finds the last stage in
        the plan and pulls records from it - this stage then pulls records from earlier
        stages in the plan as needed, and so on, until we get to a node that creates
        records to feed into the engine (usually reading a data file).
        """

        def map_operators(nodes):
            """
            We're walking the query plan telling each node where to get the data it
            needs from.
            """
            for node in nodes:
                producers = self.ingoing_edges(node)
                operator = self[node]
                if producers:
                    operator.set_producers([self[i[0]] for i in producers])
                    map_operators(i[0] for i in producers)

        # do some basic validation before we try to execute
        if not self.is_acyclic():  # pragma: no cover
            raise DatabaseError("Problem executing the query plan - it is cyclic.")

        # we get the tail of the query - the first steps
        head = list(dict.fromkeys(self.get_exit_points()))
        if len(head) != 1:  # pragma: no cover
            raise DatabaseError(f"Problem executing the query plan - it has {len(head)} heads.")

        map_operators(head)

        operator = self[head[0]]
        yield from operator.execute()

    def explain(self):
        from opteryx import operators
        from opteryx.models import Columns

        def _inner_explain(node, depth):
            if depth == 1:
                operator = self[node]
                yield {
                    "operator": operator.name,
                    "config": operator.config,
                    "depth": depth - 1,
                }
            incoming_operators = self.ingoing_edges(node)
            for operator_name in incoming_operators:
                operator = self[operator_name[0]]
                if isinstance(operator, operators.BasePlanNode):
                    yield {
                        "operator": operator.name,
                        "config": operator.config,
                        "depth": depth,
                    }
                yield from _inner_explain(operator_name[0], depth + 1)

        head = list(dict.fromkeys(self.get_exit_points()))
        if len(head) != 1:  # pragma: no cover
            raise DatabaseError(f"Problem with the plan - it has {len(head)} heads.")
        plan = list(_inner_explain(head[0], 1))

        table = pyarrow.Table.from_pylist(plan)
        table = Columns.create_table_metadata(
            table, table.num_rows, "plan", None, "calculated", "explain"
        )
        yield table



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Grouping Node

This is a SQL Query Execution Plan Node.

This performs aggregations, both of grouped and non-grouped data.

This is a greedy operator - it consumes all the data before responding.

This is built around the pyarrow table grouping functionality.
"""
import random
import time
from typing import Iterable

import numpy
import pyarrow

from opteryx.exceptions import SqlError
from opteryx.exceptions import UnsupportedSyntaxError
from opteryx.managers.expression import NodeType
from opteryx.managers.expression import evaluate_and_append
from opteryx.managers.expression import format_expression
from opteryx.managers.expression import get_all_nodes_of_type
from opteryx.models import QueryProperties
from opteryx.models.columns import Columns
from opteryx.operators import BasePlanNode

COUNT_STAR: str = "COUNT(*)"

# use the aggregators from pyarrow
AGGREGATORS = {
    "ALL": "all",
    "ANY": "any",
    "APPROXIMATE_MEDIAN": "approximate_median",
    "ARRAY_AGG": "hash_list",
    "COUNT": "count",  # counts only non nulls
    "COUNT_DISTINCT": "count_distinct",
    "DISTINCT": "distinct",  # fated
    "LIST": "hash_list",  # fated
    "MAX": "max",
    "MAXIMUM": "max",  # alias
    "MEAN": "mean",
    "AVG": "mean",  # alias
    "AVERAGE": "mean",  # alias
    "MIN": "min",
    "MINIMUM": "min",  # alias
    "MIN_MAX": "min_max",
    "ONE": "hash_one",
    "ANY_VALUE": "hash_one",
    "PRODUCT": "product",
    "STDDEV": "stddev",
    "SUM": "sum",
    "VARIANCE": "variance",
}


def _is_count_star(aggregates, groups):
    """
    Is the SELECT clause `SELECT COUNT(*)` with no GROUP BY
    """
    if len(groups) != 0:
        return False
    if len(aggregates) != 1:
        return False
    if aggregates[0].value != "COUNT":
        return False
    if aggregates[0].parameters[0].token_type != NodeType.WILDCARD:
        return False
    return True


def _count_star(morsel_promise):
    count = 0
    for morsel in morsel_promise.execute():
        count += morsel.num_rows
    table = pyarrow.Table.from_pylist([{COUNT_STAR: count}])
    table = Columns.create_table_metadata(
        table=table,
        expected_rows=1,
        name="groupby",
        table_aliases=[],
        disposition="calculated",
        path="group_by",
    )
    yield table


def _project(tables, fields):
    fields = list(dict.fromkeys(fields))
    for table in tables:
        row_count = table.num_rows
        columns = Columns(table)
        column_names = [columns.get_column_from_alias(field, only_one=True) for field in fields]
        if len(column_names) > 0:
            yield table.select(dict.fromkeys(column_names))
        else:
            # if we can't find the column, add a placeholder column
            yield pyarrow.Table.from_pydict({"*": numpy.full(row_count, 1, dtype=numpy.int8)})


def _build_aggs(aggregators, columns):
    column_map = {}
    aggs = []

    if not isinstance(aggregators, list):
        aggregators = [aggregators]

    for root in aggregators:
        for aggregator in get_all_nodes_of_type(
            root, select_nodes=(NodeType.AGGREGATOR, NodeType.COMPLEX_AGGREGATOR)
        ):
            if aggregator.token_type in (
                NodeType.AGGREGATOR,
                NodeType.COMPLEX_AGGREGATOR,
            ):
                field_node = aggregator.parameters[0]
                display_name = format_expression(field_node)
                exists = columns.get_column_from_alias(display_name)
                count_options = None

                if field_node.token_type == NodeType.WILDCARD:
                    field_name = "*"
                    # count * counts nulls
                    count_options = pyarrow.compute.CountOptions(mode="all")
                elif field_node.token_type == NodeType.IDENTIFIER:
                    field_name = columns.get_column_from_alias(field_node.value, only_one=True)
                elif field_node.token_type in (
                    NodeType.LITERAL_NUMERIC,
                    NodeType.LITERAL_BOOLEAN,
                    NodeType.LITERAL_VARCHAR,
                ):
                    field_name = str(field_node.value)
                elif len(exists) > 0:
                    field_name = exists[0]
                else:
                    display_name = format_expression(field_node)
                    raise SqlError(
                        f"Invalid identifier or literal provided in aggregator function `{display_name}`"
                    )
                function = AGGREGATORS.get(aggregator.value)
                if aggregator.value == "ARRAY_AGG":
                    # if the array agg is distinct, base off that function instead
                    if aggregator.parameters[1]:
                        function = "distinct"
                aggs.append((field_name, function, count_options))
                column_map[
                    format_expression(aggregator)
                    #                    f"{aggregator.value.upper()}({display_field})"
                ] = f"{field_name}_{function}".replace("_hash_", "_")

    return column_map, aggs


def _non_group_aggregates(aggregates, table, columns):
    """
    If we're not doing a group by, we're just doing aggregations, the pyarrow
    functionality for aggregate doesn't work. So we do the calculation, it's
    relatively straightforward as it's the entire table we're summarizing.
    """

    result = {}

    for aggregate in aggregates:
        if aggregate.token_type in (NodeType.AGGREGATOR, NodeType.COMPLEX_AGGREGATOR):
            column_node = aggregate.parameters[0]
            if column_node.token_type == NodeType.LITERAL_NUMERIC:
                raw_column_values = numpy.full(
                    table.num_rows, column_node.value, dtype=numpy.float64
                )
                mapped_column_name = str(column_node.value)
            elif (
                aggregate.value == "COUNT"
                and aggregate.parameters[0].token_type == NodeType.WILDCARD
            ):
                result["COUNT(*)"] = table.num_rows
                continue
            else:
                column_name = format_expression(aggregate.parameters[0])
                mapped_column_name = columns.get_column_from_alias(column_name, only_one=True)
                raw_column_values = table[mapped_column_name].to_numpy()
            aggregate_function_name = AGGREGATORS[aggregate.value]
            # this maps a string which is the function name to that function on the
            # pyarrow.compute module
            if not hasattr(pyarrow.compute, aggregate_function_name):
                raise UnsupportedSyntaxError(
                    f"Aggregate `{aggregate.value}` can only be used with GROUP BY"
                )
            aggregate_function = getattr(pyarrow.compute, aggregate_function_name)
            aggregate_column_value = aggregate_function(raw_column_values).as_py()
            aggregate_column_name = f"{mapped_column_name}_{aggregate_function_name}"
            result[aggregate_column_name] = aggregate_column_value

    return pyarrow.Table.from_pylist([result])


def _extract_functions(aggregates):
    # extract any inner evaluations, like the IIF in SUM(IIF(x, 1, 0))

    all_evaluatable_nodes = get_all_nodes_of_type(
        aggregates,
        select_nodes=(
            NodeType.FUNCTION,
            NodeType.BINARY_OPERATOR,
            NodeType.COMPARISON_OPERATOR,
        ),
    )

    evaluatable_nodes = []
    for node in all_evaluatable_nodes:
        aggregators = get_all_nodes_of_type(
            node, select_nodes=(NodeType.AGGREGATOR, NodeType.COMPLEX_AGGREGATOR)
        )
        if len(aggregators) == 0:
            evaluatable_nodes.append(node)

    return evaluatable_nodes


class AggregateNode(BasePlanNode):
    def __init__(self, properties: QueryProperties, **config):
        super().__init__(properties=properties)

        self._aggregates = config.get("aggregates", [])
        if any(node.token_type == NodeType.WILDCARD for node in self._aggregates):
            raise SqlError("Cannot select `*` with `GROUP BY` clause.")

        self._groups = []
        for group in config.get("groups", []):
            # skip nulls
            if group is None:
                continue
            # handle numeric indexes - GROUP BY 1
            if group.token_type == NodeType.LITERAL_NUMERIC:
                # references are natural numbers, so -1 for zero-based
                group = self._aggregates[int(group.value) - 1]
                if group.token_type not in (NodeType.IDENTIFIER, NodeType.FUNCTION):
                    raise SqlError(
                        "When using a positional reference, GROUP BY must be by an IDENTIFIER or FUNCTION only"
                    )
            self._groups.append(group)

    @property
    def config(self):  # pragma: no cover
        return str(self._aggregates)

    @property
    def greedy(self):  # pragma: no cover
        return True

    @property
    def name(self):  # pragma: no cover
        return "Aggregation"

    def execute(self) -> Iterable:
        if len(self._producers) != 1:  # pragma: no cover
            raise SqlError(f"{self.name} on expects a single producer")

        morsels = self._producers[0]  # type:ignore
        if isinstance(morsels, pyarrow.Table):
            morsels = (morsels,)

        if _is_count_star(self._aggregates, self._groups):
            yield from _count_star(morsel_promise=morsels)
            return

        # get all the columns anywhere in the groups or aggregates
        all_identifiers = [
            node.value
            for node in get_all_nodes_of_type(
                self._groups + self._aggregates, select_nodes=(NodeType.IDENTIFIER,)
            )
        ]
        all_identifiers = list(dict.fromkeys(all_identifiers))

        # merge all the morsels together into one table, selecting only the columns
        # we're pretty sure we're going to use - this will fail for datasets
        # larger than memory
        table = pyarrow.concat_tables(_project(morsels.execute(), all_identifiers), promote=True)

        # Get any functions we need to execute before aggregating
        evaluatable_nodes = _extract_functions(self._aggregates)

        # Allow grouping by functions by evaluating them first
        start_time = time.time_ns()
        columns, _, table = evaluate_and_append(evaluatable_nodes, table)
        columns, self._groups, table = evaluate_and_append(self._groups, table)

        # Add a "*" column, this is an int because when a bool it miscounts
        if "*" not in table.column_names:
            table = table.append_column(
                "*", [numpy.full(shape=table.num_rows, fill_value=1, dtype=numpy.int8)]
            )
        self.statistics.time_evaluating += time.time_ns() - start_time

        # Extract any literal columns, we need to add these so we can group and/or
        # aggregate by them (e.g. SELECT SUM(4) FROM table;)
        all_literals = [
            node.value
            for node in get_all_nodes_of_type(
                self._groups + self._aggregates,
                select_nodes=(
                    NodeType.LITERAL_BOOLEAN,
                    NodeType.LITERAL_NUMERIC,
                    NodeType.LITERAL_VARCHAR,
                ),
            )
        ]
        all_literals = list(dict.fromkeys(all_literals))
        all_literals = [a for a in all_literals if str(a) not in table.column_names]
        for literal in all_literals:
            table = table.append_column(
                str(literal), [numpy.full(shape=table.num_rows, fill_value=literal)]
            )
            columns.add_column(str(literal))

        start_time = time.time_ns()
        # GROUP BY columns are deduplicated #870
        group_by_columns = list(
            dict.fromkeys(
                columns.get_column_from_alias(group.value, only_one=True) for group in self._groups
            )
        )

        column_map, aggs = _build_aggs(self._aggregates, columns)

        # we're not a group_by - we're aggregating without grouping
        if len(group_by_columns) == 0:
            groups = _non_group_aggregates(self._aggregates, table, columns)
            del table
        else:
            table = table.combine_chunks()
            groups = table.group_by(group_by_columns)
            groups = groups.aggregate(aggs)

        # do the secondary activities for ARRAY_AGG
        for node in get_all_nodes_of_type(
            self._aggregates, select_nodes=(NodeType.COMPLEX_AGGREGATOR,)
        ):
            if node.value == "ARRAY_AGG":
                _, _, order, limit = node.parameters
                if order or limit:
                    # rip the column out of the table
                    column_name = column_map[format_expression(node)]
                    column_def = groups.field(column_name)  # this is used
                    column = groups.column(column_name).to_pylist()
                    groups = groups.drop([column_name])
                    # order
                    if order:
                        pass
                    if limit:
                        column = [c[:limit] for c in column]
                    # put the new column into the table
                    groups = groups.append_column(column_def, [column])

        # name the aggregate fields and add them to the Columns data
        for friendly_name, agg_name in column_map.items():
            # randomize the column name to prevent collisions, particularly on joined
            # or chained aggregations
            new_column_name = hex(random.getrandbits(64))
            groups = groups.rename_columns(
                [col if col != agg_name else new_column_name for col in groups.column_names]
            )

            columns.add_column(new_column_name)
            columns.set_preferred_name(new_column_name, friendly_name)
            # if we have an alias for this column, add it to the metadata
            aliases = [
                agg.alias for agg in self._aggregates if friendly_name == format_expression(agg)
            ]
            aliases.append(agg_name)
            for alias in aliases:
                if alias:
                    columns.add_alias(new_column_name, alias)

        # apply the metadata to the resultant dataset
        groups = columns.apply(groups)

        self.statistics.time_aggregating += time.time_ns() - start_time

        yield groups



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Show Columns Node

This is a SQL Query Execution Plan Node.

Gives information about a dataset's columns
"""
from functools import reduce
from typing import Iterable

import numpy
import orjson
import pyarrow
from numpy import nan
from numpy import nanmax
from numpy import nanmin

from opteryx.constants.attribute_types import OPTERYX_TYPES
from opteryx.constants.attribute_types import determine_type
from opteryx.exceptions import SqlError
from opteryx.models import Columns
from opteryx.models import QueryProperties
from opteryx.operators import BasePlanNode

MAX_COLLECTOR: int = 17
MAX_VARCHAR_SIZE: int = 64  # long strings tend to lose meaning
MAX_DATA_SIZE: int = 100 * 1024 * 1024


def _to_unix_epoch(date):
    timestamp = date.value
    if timestamp is None:
        return numpy.nan
    return timestamp / 1e6


def _simple_collector(morsel):
    """
    Collect the very summary type information only, we read only a single morsel to do
    this so it's pretty quick - helpful if you want to know what fields are available
    programatically.
    """
    columns = Columns(morsel)

    buffer = []
    for column in morsel.column_names:
        column_data = morsel.column(column)
        _type = determine_type(str(column_data.type))
        new_row = {"name": columns.get_preferred_name(column), "type": _type}
        buffer.append(new_row)

    table = pyarrow.Table.from_pylist(buffer)
    table = Columns.create_table_metadata(
        table=table,
        expected_rows=len(buffer),
        name="show_columns",
        table_aliases=[],
        disposition="calculated",
        path="show_columns",
    )
    return table


def _full_collector(morsels):
    """
    Collect basic count information about columns, to do this we read the entire
    dataset.
    """

    empty_profile = orjson.dumps(
        {
            "name": None,
            "type": [],
            "count": 0,
            "min": None,
            "max": None,
            "missing": 0,
        }
    )

    columns = None
    profile_collector: dict = {}

    for morsel in morsels:
        if columns is None:
            columns = Columns(morsel)

        for column in morsel.column_names:
            column_data = morsel.column(column)
            profile = profile_collector.get(column, orjson.loads(empty_profile))
            _type = determine_type(str(column_data.type))
            if _type not in profile["type"]:
                profile["type"].append(_type)

            profile["count"] += len(column_data)

            # calculate the missing count more robustly
            missing = reduce(
                lambda x, y: x + 1,
                (i for i in column_data if i in (None, nan) or not i.is_valid),
                0,
            )
            profile["missing"] += missing

            if _type == OPTERYX_TYPES.NUMERIC:
                if profile["min"]:
                    profile["min"] = min(profile["min"], nanmin(column_data))
                    profile["max"] = max(profile["max"], nanmax(column_data))
                else:
                    profile["min"] = nanmin(column_data)
                    profile["max"] = nanmax(column_data)

            profile_collector[column] = profile

    buffer = []

    for column, profile in profile_collector.items():
        profile["name"] = columns.get_preferred_name(column)
        profile["type"] = ", ".join(profile["type"])
        buffer.append(profile)

    table = pyarrow.Table.from_pylist(buffer)
    table = Columns.create_table_metadata(
        table=table,
        expected_rows=len(buffer),
        name="show_columns",
        table_aliases=[],
        disposition="calculated",
        path="show_columns",
    )
    return table


def increment(dic: dict, value):
    if value in dic:
        dic[value] += 1
    else:
        dic[value] = 1


def _extended_collector(morsels):
    """
    Collect summary statistics about each column
    """
    import orso
    from orso import converters

    from opteryx import utils

    rows, schema = converters.from_arrow(utils.arrow.rename_columns(morsels))
    df = orso.DataFrame(rows=rows, schema=schema)
    table = df.profile.arrow()
    table = Columns.create_table_metadata(
        table=table,
        expected_rows=table.num_rows,
        name="show_columns",
        table_aliases=[],
        disposition="calculated",
        path="show_columns",
    )
    return table


class ShowColumnsNode(BasePlanNode):
    def __init__(self, properties: QueryProperties, **config):
        super().__init__(properties=properties)
        self._full = config.get("full")
        self._extended = config.get("extended")

    @property
    def name(self):  # pragma: no cover
        return "Show Columns"

    @property
    def config(self):  # pragma: no cover
        return ""

    def execute(self) -> Iterable:
        # TODO: [TARCHIA] - use the metastore to get the column statisitcs

        if len(self._producers) != 1:  # pragma: no cover
            raise SqlError(f"{self.name} on expects a single producer")

        morsels = self._producers[0]  # type:ignore

        if morsels is None:
            return None

        if not (self._full or self._extended):
            # if it's not full or extended, do just get the list of columns and their
            # types
            yield _simple_collector(next(morsels.execute()))
            return

        if self._full and not self._extended:
            # we're going to read the full table, so we can count stuff
            yield _full_collector(morsels.execute())
            return

        if self._extended:
            # get everything we can reasonable get
            yield _extended_collector(morsels.execute())
            return



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Show Variables Node

This is a SQL Query Execution Plan Node.
"""
from typing import Iterable

import pyarrow

from opteryx import __version__
from opteryx.exceptions import SqlError
from opteryx.models import Columns
from opteryx.operators import BasePlanNode

SYSTEM_VARIABLES = [{"variable_name": "version", "value": __version__}]


class ShowVariablesNode(BasePlanNode):
    @property
    def name(self):  # pragma: no cover
        return "Show Variables"

    @property
    def config(self):  # pragma: no cover
        return ""

    def execute(self) -> Iterable:
        buffer = SYSTEM_VARIABLES

        for variable, value in self.properties.variables.items():
            buffer.append({"variable_name": variable, "value": str(value.value)})

        table = pyarrow.Table.from_pylist(buffer)
        table = Columns.create_table_metadata(
            table=table,
            expected_rows=len(buffer),
            name="show_variables",
            table_aliases=[],
            disposition="calculated",
            path="show_variable",
        )

        yield table
        return



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Optimization Rule - Eliminite Fixed-Outcome Function Evaluations

Type: Heuristic
Goal: Reduce complexity
"""
from opteryx import operators
from opteryx.functions import date_functions
from opteryx.functions import get_version
from opteryx.functions import number_functions
from opteryx.managers.expression import ExpressionTreeNode
from opteryx.managers.expression import NodeType

FIXED_OUTCOME_FUNCTIONS = {
    # TEMPORAL
    "CURRENT_TIME": (date_functions.get_now, NodeType.LITERAL_TIMESTAMP),
    "NOW": (date_functions.get_now, NodeType.LITERAL_TIMESTAMP),
    "UTC_TIMESTAMP": (date_functions.get_now, NodeType.LITERAL_TIMESTAMP),
    "CURRENT_DATE": (date_functions.get_today, NodeType.LITERAL_TIMESTAMP),
    "TODAY": (date_functions.get_today, NodeType.LITERAL_TIMESTAMP),
    "TIME": (date_functions.get_time, NodeType.LITERAL_TIMESTAMP),
    "YESTERDAY": (date_functions.get_yesterday, NodeType.LITERAL_TIMESTAMP),
    # SYSTEM
    "VERSION": (get_version, NodeType.LITERAL_VARCHAR),
    # MATHEMATICAL
    "E": (number_functions.e, NodeType.LITERAL_NUMERIC),
    "PI": (number_functions.pi, NodeType.LITERAL_NUMERIC),
    "PHI": (number_functions.phi, NodeType.LITERAL_NUMERIC),
}


def eliminate_fixed_function_evaluations(plan, properties):
    """
    This action aims to execute functions once and replace the outcome as a literal in
    the query, where the function evaluation will be the same for the query, for
    example `current_time`
    """

    def update_expression_tree(node):
        """
        Walk a expression tree collecting all the nodes of a specified type.
        """

        # this is the main work of this action
        if node.token_type == NodeType.FUNCTION and node.value in FIXED_OUTCOME_FUNCTIONS:
            function, node_type = FIXED_OUTCOME_FUNCTIONS[node.value]
            return ExpressionTreeNode(node_type, value=function(), alias=node.alias)

        # walk the tree
        node.left = None if node.left is None else update_expression_tree(node.left)
        node.centre = None if node.centre is None else update_expression_tree(node.centre)
        node.right = None if node.right is None else update_expression_tree(node.right)

        if node.parameters:
            node.parameters = [
                parameter
                if not isinstance(parameter, ExpressionTreeNode)
                else update_expression_tree(parameter)
                for parameter in node.parameters
            ]

        return node

    # find the in-scope nodes (WHERE AND HAVING)
    selection_nodes = plan.get_nodes_of_type(operators.SelectionNode)

    # killer questions - if any aren't met, bail
    if selection_nodes is None:
        return plan

    # HAVING and WHERE are selection nodes
    for nid in selection_nodes:
        # get the node from the node_id
        operator = plan[nid]
        operator.filter = update_expression_tree(operator.filter)

    return plan



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random

import numpy
from pyarrow import compute


def pi():
    return 3.14159265358979323846264338327950288419716939937510


def phi():
    """the golden ratio"""
    return 1.61803398874989484820458683436563811772030917980576


def e():
    """eulers number"""
    return 2.71828182845904523536028747135266249775724709369995


def round(*args):
    if len(args) == 1:
        return compute.round(args[0])
    # the second parameter is a fixed value
    return compute.round(args[0], args[1][0])  # [#325]


def random_number(size):
    return numpy.random.uniform(size=size)


def random_normal(size):
    from numpy.random import default_rng

    rng = default_rng()
    return rng.standard_normal(size)


def random_string(width):
    # this is roughly twice as fast the the previous implementation
    # a tuple is slightly faster than a string, don't use a list
    width = int(width)
    # fmt:off
    alphabet = ('a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '_', '/')
    # ftm:on
    return "".join([alphabet[random.getrandbits(6)] for i in range(width)])



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains support functions for working with PyArrow
"""

import pyarrow
from orjson import dumps
from orjson import loads

INTERNAL_BATCH_SIZE = 500


def limit_records(morsels, limit):
    """
    Cycle over an iterable of morsels, limiting the response to a given
    number of records.
    """
    row_count = 0
    result_set = []
    morsels_iterator = iter(morsels)

    morsel = None
    while row_count < limit:
        try:
            morsel = next(morsels_iterator)
        except StopIteration:
            break

        if morsel.num_rows > 0:
            if row_count + morsel.num_rows <= limit:
                # append whole morsel to result_set
                result_set.append(morsel)
                row_count += morsel.num_rows
            else:
                # slice morsel to only append needed rows to result_set
                num_rows_needed = limit - row_count
                result_set.append(morsel.slice(offset=0, length=num_rows_needed))
                row_count = limit

    if len(result_set) == 0:
        if morsel is None:
            morsel = next(morsels_iterator, None)
        if morsel is None:
            return None
        return pyarrow.Table.from_batches([], schema=morsel.schema)
    else:
        return pyarrow.concat_tables(result_set, promote=True)


def rename_columns(morsels):
    """rename columns to their preferred names"""
    from opteryx.models import Columns

    columns = None
    morsels = iter(morsels)
    for morsel in morsels:
        if morsel is None:
            break
        if columns is None and morsel is not None:
            columns = Columns(morsel)
            preferred_names = columns.preferred_column_names
            column_names = []
            for col in morsel.column_names:
                column_names.append([c for a, c in preferred_names if a == col][0])
        if column_names and morsel is not None:
            yield morsel.rename_columns(column_names)


# Adapted from:
# https://stackoverflow.com/questions/55546027/how-to-assign-arbitrary-metadata-to-pyarrow-table-parquet-columns


def set_metadata(table, table_metadata=None, column_metadata=None):
    """
    Store table-level metadata as json-encoded byte strings.

    Table-level metadata is stored in the table's schema.

    parameters:
        table: pyarrow.Table
            The table to store metadata in
        col_meta: dict
            A json-serializable dictionary with column metadata in the form
            {
                'column_1': {'some': 'data', 'value': 1},
                'column_2': {'more': 'stuff', 'values': [1,2,3]}
            }
        tbl_meta: dict
            A json-serializable dictionary with table-level metadata.
    """

    # Create updated column fields with new metadata
    if table_metadata or column_metadata:
        fields = []
        for name in table.schema.names:
            col = table.field(name)
            if col.name in column_metadata:
                # Get updated column metadata
                metadata = col.metadata or {}
                for k, v in column_metadata[name].items():
                    if isinstance(k, str):
                        k = k.encode()
                    metadata[k] = dumps(v)
                # Update field with updated metadata
                fields.append(col.with_metadata(metadata))
            else:
                fields.append(col)

        # Get updated table metadata
        tbl_metadata = table.schema.metadata or {}
        if table_metadata:
            for k, v in table_metadata.items():
                if isinstance(v, bytes):
                    tbl_metadata[k] = v
                else:
                    tbl_metadata[k] = dumps(v)

        # Create new schema with updated table metadata
        schema = pyarrow.schema(fields, metadata=tbl_metadata)
        # With updated schema build new table (shouldn't copy data)
        table = table.cast(schema)

    return table


def _decode_metadata(metadata):
    """
    Arrow stores metadata keys and values as bytes. We store "arbitrary" data as
    json-encoded strings (utf-8), which are here decoded into normal dict.
    """

    if not metadata:
        # None or {} are not decoded
        return {}

    decoded = {}
    for key, value in metadata.items():
        key = key.decode("utf-8")
        val = loads(value)
        decoded[key] = val
    return decoded


def table_metadata(tbl):
    """Get table metadata as dict."""
    return _decode_metadata(tbl.schema.metadata)


def column_metadata(tbl):
    """Get column metadata as dict."""
    return {col: _decode_metadata(tbl.field(col).metadata) for col in tbl.schema.names}


def coerce_columns(table, column_names):
    """convert numeric types to a common type to allow comparisons"""
    # get the column we're coercing
    my_schema = table.schema

    if not isinstance(column_names, list):
        column_names = [column_names]

    for column_name in column_names:
        index = table.column_names.index(column_name)
        column = my_schema.field(column_name)

        # if it's numeric, and not already the type we want, convert it
        if str(column.type) in ("int64", "double"):
            column = column.with_type(pyarrow.float64())
            my_schema = my_schema.set(index, pyarrow.field(column_name, pyarrow.float64()))
            table = table.cast(target_schema=my_schema)

    return table



"""
travers

(C) 2023 Justin Joyce.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from pathlib import Path
from typing import List
from typing import Optional
from typing import Tuple

import orjson

from opteryx.exceptions import MissingDependencyError


class Graph(object):
    """
    Graph object, optimized for traversal.

    Edges are stored in a dictionary, the key is the source node to speed up
    finding outgoing edges. The Edges only have three pieces of data:
        - the source node (the key)
        - the target node
        - the relationship
    The target and the relationship are stored as a tuple, the edge dictionary
    stores lists of tuples.

    """

    __slots__ = ("_nodes", "_edges")

    def __init__(self):
        """
        Directed Graph.
        """
        self._nodes = {}
        self._edges = {}

    def _make_a_list(self, obj):
        """internal helper method"""
        if isinstance(obj, list):
            return obj
        return [obj]

    def save(self, graph_path):  # pragma: nocover
        """
        Persist a graph to storage. It saves nodes and edges to separate files.

        Parameters:
            graph_path: string
                The folder ?to save the node and edge files to
        """
        path = Path(graph_path)
        path.mkdir(exist_ok=True)

        with open(path / "edges.jsonl", "wb") as edge_file:
            for source, target, relationship in self.edges():
                edge_record = {
                    "source": source,
                    "target": target,
                    "relationship": relationship,
                }
                edge_file.write(orjson.dumps(edge_record) + b"\n")
        with open(path / "nodes.jsonl", "wb") as node_file:
            for nid, attr in self.nodes(data=True):
                node_file.write(orjson.dumps({"nid": nid, "attributes": attr}) + b"\n")

    def add_edge(self, source: str, target: str, relationship: Optional[str] = None):
        """
        Add edge to the graph

        Note:
            This does not create an edge if either node does not already exist.
            This does not create an edge if either node is None.

        Parameters:
            source: string
                The source node
            target: string
                The target node
            relationship: string
                The relationship between the source and target nodes
        """
        if source is None or target is None:
            print("Trying to create edge with undefined nodes")
            return False

        if source not in self._edges:
            targets = []
        else:
            targets = self._edges[source]

        targets.append(
            (
                target,
                relationship,
            )
        )
        self._edges[source] = list(set(targets))
        return True

    def add_node(self, nid: str, node):
        """
        Add node to the graph

        Parameters:
            nid: string
                The Node ID for the node (unique)
            attributes: dictionary (optional)
                The attributes of the node
        """
        self._nodes[nid] = node

    def nodes(self, data=False):
        """
        The nodes which comprise the graph

        Parameters:
            data: boolean (optional)
                if True return the details of the nodes, if False just return
                the list of node IDs

        Returns:
            List
        """
        if data:
            return self._nodes.items()
        return list(self._nodes.keys())

    def edges(self):
        """
        The edges which comprise the graph

        Returns:
            Generator of Tuples of (Source, Target and Relationship)
        """
        for source, records in self._edges.items():
            yield from ((source, target, relationship) for target, relationship in records)

    def breadth_first_search(self, source: str, depth: int = 100):  # pragma: nocover
        """
        Search a tree for nodes we can walk to from a given node.

        Parameters:
            source: string
                The node to walk from
            depth: integer
                The maximum distance to walk from source
        Returns:
        """
        # This uses a variation of the algorith used by NetworkX optimized for
        # the travers data structures.
        #
        # https://networkx.org/documentation/networkx-1.10/_modules/networkx/algorithms/traversal/breadth_first_search.html#bfs_tree

        from collections import deque

        distance = 0

        visited = set([source])
        queue = deque(
            [
                (
                    source,
                    distance,
                    self.outgoing_edges(source),
                )
            ]
        )

        new_edges = []

        while queue:
            parent, node_distance, children = queue[0]
            if node_distance < depth:
                for child in children:
                    s, t, r = child
                    new_edges.append(child)
                    if t not in visited:
                        visited.add(t)
                        queue.append(
                            (
                                t,
                                node_distance + 1,
                                self.outgoing_edges(t),
                            )
                        )
            queue.popleft()
        return new_edges

    def depth_first_search(
        self, node: Optional[str] = None, visited: Optional[set] = None, depth: int = 0
    ):
        """
        Returns a nested dictionary representing the graph as a tree
        """
        if node is None:
            node = self.get_exit_points()[0]

        if visited is None:
            visited = set()

        visited.add(node)

        tree: dict = {
            "type": str(self[node].node_type),
            "name": node,
            "depth": depth,
            "children": [],
        }

        for neighbor, _, relationship in self.ingoing_edges(node):
            if neighbor not in visited:
                child = self.depth_first_search(neighbor, visited, depth + 1)
                child["relationship"] = relationship
                tree["children"].append(child)  # type:ignore

        return tree

    def outgoing_edges(self, source) -> List[Tuple]:
        """
        Get the list of edges traversable from a given node.

        Parameters:
            source: string
                The node to get the outgoing edges for

        Returns:
            Set of Tuples (Source, Target, Relationship)
        """
        return [(source, t, r) for t, r in self._edges.get(source, [])]

    def ingoing_edges(self, target) -> List[Tuple]:
        """
        Get the list of edges which can traverse to a given node.

        Parameters:
            target: string
                The node to get the incoming edges for

        Returns:
            Set of Tuples (Source, Target, Relationship)
        """
        return [(s, t, r) for s, t, r in self.edges() if t == target]

    def is_acyclic(self):
        """
        Test if the Graph is acyclic
        """
        # cycle over the graph removing a layer of exits each cycle
        # if we have nodes but no exists, we're cyclic

        # rebuild the edge information
        my_edges = list(self.edges())

        while len(my_edges) > 0:
            # find all of the exits
            sources = {source for source, target, direction in my_edges}
            exits = {target for source, target, direction in my_edges if target not in sources}

            if len(exits) == 0:
                return False

            # remove the exits
            new_edges = [
                (source, target, direction)
                for source, target, direction in my_edges
                if target not in exits
            ]
            my_edges = new_edges
        return True

    def get_entry_points(self):
        """
        Get nodes in the Graph with no incoming edges.
        """
        if len(self._nodes) == 1:
            return list(self._nodes.keys())
        targets = {target for source, target, direction in self.edges()}
        retval = (source for source, target, direction in self.edges() if source not in targets)
        return sorted(retval)

    def get_exit_points(self):
        """
        Get nodes in the Graph with no outgoing edges.
        """
        if len(self._nodes) == 1:  # pragma: no cover
            return list(self._nodes.keys())
        sources = self._edges.keys()
        retval = (target for source, target, direction in self.edges() if target not in sources)
        return sorted(retval)

    def remove_node(self, nid, heal: bool = False):
        """
        Remove a node.

        Parameters:
            heal: boolean
                Join the incoming and outgoing connections for the removed node
                to each other to keep the Graph intact
        """

        # remove the node
        self._nodes.pop(nid, None)

        if heal:
            # link the nodes each side of the node being removed
            out_going = self.outgoing_edges(nid)
            in_coming = self.ingoing_edges(nid)

            # remove edges where the node is the source
            if nid in self._edges:
                del self._edges[nid]

            # remove the edges where the node is the target
            for source, records in self._edges.items():
                self._edges[source] = [
                    (target, relationship) for target, relationship in records if target != nid
                ]
            self._edges = {k: v for k, v in self._edges.items() if len(v) > 0}

            # wire up the old incoming and outgoing nodes, cartesian style
            for out_nid in out_going:
                for in_nid in in_coming:
                    self.add_edge(in_nid[0], out_nid[1], in_nid[1])  # type:ignore

    def insert_node_before(self, nid, node, before_nid):
        """rewrite the plan putting the new node before a given node"""
        # add the new node to the plan
        self.add_node(nid, node)
        # change all the edges that were going into the old nid to the new one

        for source, records in self._edges.items():
            new_records = []
            for target, relationship in records:
                if target != before_nid:
                    new_records.append(
                        (
                            target,
                            relationship,
                        )
                    )
                else:
                    new_records.append(
                        (
                            nid,
                            relationship,
                        )
                    )
                self._edges[source] = new_records
        # add an edge from the new nid to the old one
        self.add_edge(nid, before_nid)

    def insert_node_after(self, nid, node, after_nid):
        """rewrite the plan putting the new node after a given node"""
        # add the new node to the plan
        self.add_node(nid, node)
        # change all the edges that were coming from the old nid to the new one
        if after_nid in self._edges:
            self._edges[nid] = self._edges.pop(after_nid)
        # add an edge from the new nid to the old one
        self.add_edge(after_nid, nid)

    def copy(self):  # pragma: nocover
        g = Graph()
        g._nodes = self._nodes.copy()
        g._edges = self._edges.copy()
        return g

    def to_networkx(self):  # pragma: nocover
        """
        Convert a travers graph to a NetworkX graph
        """
        try:
            import networkx as nx  # type:ignore
        except ImportError as err:  # pragma: no cover
            raise MissingDependencyError(err.name) from err

        g = nx.DiGraph()
        for s, t, r in self.edges():
            g.add_edge(s, t, relationship=r)
        for node, attribs in self.nodes(True):
            g.add_node(node, **attribs)
        return g

    def epitomize(self):  # pragma: nocover
        """
        Summarize a Graph by reducing to only the node_types and relationships
        """
        g = Graph()
        for s, t, r in self.edges():
            node1 = self[s]
            node2 = self[t]
            if node1 and node2:
                g.add_edge(node1.get("node_type"), node2.get("node_type"), r)
            if node1:
                g.add_node(node1.get("node_type"), {"node_type": node1.get("node_type")})
            if node2:
                g.add_node(node2.get("node_type"), {"node_type": node2.get("node_type")})
        return g

    def __repr__(self):
        return f"Graph - {len(list(self.nodes()))} nodes, {len(list(self.edges()))} edges"

    def __len__(self):
        return len(list(self.nodes()))

    def __getitem__(self, nid):
        return self._nodes.get(nid, None)

    def __add__(self, other):
        self._edges.update(other._edges)
        self._nodes.update(other._nodes)
        return self

    # adapted from https://stackoverflow.com/questions/9727673/list-directory-tree-structure-in-python
    def _tree(self, node, prefix=""):
        space = "   "
        branch = " │ "
        tee = " ├─"
        last = " └─"

        contents = [node for node, _, _ in self.ingoing_edges(node)]
        # contents each get pointers that are ├── with a final └── :
        pointers = [tee] * (len(contents) - 1) + [last]
        for pointer, child_node in zip(pointers, contents):
            label = str(self[child_node])
            yield prefix + pointer + label
            if len(self.ingoing_edges(node)) > 0:  # extend the prefix and recurse:
                extension = branch if pointer == tee else space
                # i.e. space because last, └── , above so no more |
                yield from self._tree(child_node, prefix=prefix + extension)

    def draw(self):
        tree = ""
        for entry in self.get_exit_points():
            label = str(self[entry].node_type)
            tree += label + "\n"
            t = self._tree(entry, "")
            tree += "\n".join(t) + "\n"
        return tree



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Inner Reader for SQL stores

This currently isn't a base class because we're assuming a standard
functionality of SQL engines.

This relies on SQLAlchemy
"""
import typing
from functools import lru_cache

import pyarrow

from opteryx import config
from opteryx.connectors.capabilities import PredicatePushable
from opteryx.exceptions import MissingDependencyError


class BaseSQLStorageAdapter:  # this is used by the SHOW STORES statement
    pass


def _write_predicate(predicate):
    column, operator, literal = predicate

    operator_map = {"==": "="}
    operator = operator_map.get(operator, operator)

    if isinstance(literal, str):
        literal = "'" + literal.replace("'", "''") + "'"

    return f"{column} {operator} {literal}"


@lru_cache(8)
def _get_engine(connection):
    """take advantage of pooling"""
    from sqlalchemy import create_engine

    return create_engine(connection)


class SqlConnector(BaseSQLStorageAdapter, PredicatePushable):
    __mode__ = "SQL"

    def __init__(
        self, prefix: str = "", remove_prefix: bool = False, connection: str = None
    ) -> None:
        super(BaseSQLStorageAdapter, self).__init__()
        super(PredicatePushable, self).__init__()
        # we're just testing we can import here
        try:
            from sqlalchemy import create_engine
        except ImportError as err:  # pragma: nocover
            raise MissingDependencyError(err.name) from err

        self._connection = connection

        self._remove_prefix = remove_prefix
        self._prefix = prefix

    def read_records(
        self,
        dataset,
        selection: typing.Union[list, None] = None,
        morsel_size: typing.Union[int, None] = None,
    ):  # pragma: no cover
        """
        Return a morsel of documents
        """
        from sqlalchemy import text

        from opteryx.third_party.query_builder import Query

        if morsel_size is None:
            morsel_size = config.MORSEL_SIZE
        chunk_size = 500

        queried_relation = dataset
        if self._remove_prefix:
            if dataset.startswith(f"{self._prefix}."):
                queried_relation = dataset[len(self._prefix) + 1 :]

        # we're using a query builder to prevent hand-crafting SQL
        query_builder = Query()
        query_builder.FROM(queried_relation)
        if selection is None:
            query_builder.SELECT("*")
        else:
            query_builder.SELECT(*selection)

        for predicate in self._predicates:
            query_builder.WHERE(_write_predicate(predicate))

        engine = _get_engine(self._connection)
        with engine.connect() as conn:
            result = conn.execute(text(str(query_builder)))

            batch = result.fetchmany(chunk_size)
            while batch:
                morsel = pyarrow.Table.from_pylist([b._asdict() for b in batch])
                # from 500 records, estimate the number of records to fill the morsel size
                if chunk_size == 500 and morsel.nbytes > 0:
                    chunk_size = int(morsel_size // (morsel.nbytes / 500))
                yield morsel
                batch = result.fetchmany(chunk_size)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This builds a logical plan can resolve the query from the user.

This doesn't attempt to do optimization, this just build a convenient plan which will
respond to the query correctly.

The effective order of operations must be:

    01. FROM
    02. JOIN
    03. WHERE
    04. GROUP BY
    05. HAVING
    06. SELECT
    07. DISTINCT
    08. ORDER BY
    09. OFFSET
    10. LIMIT

So we just build it in that order.
"""
from opteryx.components.logical_planner import queries
from opteryx.exceptions import UnsupportedSyntaxError


def create_logical_plan(ast, properties):
    query_type = next(iter(ast))
    builder = queries.QUERY_BUILDER.get(query_type)
    if builder is None:
        raise UnsupportedSyntaxError(f"Statement not supported `{query_type}`")
    return builder(ast, properties)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import random

import numpy
from orso.cityhash import CityHash64


def hasher(vals):
    """
    Quickly hash a list of string or numerics (i.e. intended for join hash table)

    This is roughly 2x faster than the previous implementation for lists of strings.

    Do note though, if you're micro-optimizing, this is faster to create but is
    slower for some Python functions to handle the result of, like 'sorted'.
    """
    if numpy.issubdtype(vals.dtype, numpy.character):
        return numpy.array([CityHash64(s.encode()) for s in vals], numpy.uint64)
    return vals


def peek(iterable):  # type:ignore
    """
    peek an item off a generator
    """
    iter1, iter2 = itertools.tee(iterable)
    try:
        first = next(iter1)
    except StopIteration:
        return None, iter([])
    else:
        return first, iter2


def fuzzy_search(name, candidates):
    """
    Find closest match using a Levenshtein Distance variation
    """
    from opteryx.third_party.mbleven import compare

    best_match_column = None
    best_match_score = 100

    for candidate in candidates:
        my_dist = compare(candidate, name)
        if 0 <= my_dist < best_match_score:
            best_match_score = my_dist
            best_match_column = candidate

    return best_match_column


def random_int() -> int:
    """
    Select a random integer (32bit)
    """
    return random.getrandbits(32)


def random_string(width: int = 16):
    num_chars = ((width + 1) >> 1) << 3  # Convert length to number of bits
    rand_bytes = random.getrandbits(num_chars)  # Generate random bytes
    rand_hex = hex(rand_bytes)[
        2 : width + 2
    ]  # Convert to hex string and truncate to desired length
    return rand_hex


def is_arm():
    """am I running on an ARM CPU?"""
    import platform

    return platform.machine() in ("armv7l", "aarch64")
